import random
import os
from typing import Literal, List

from adalflow.utils.lazy_import import safe_import, OptionalPackages


from adalflow.utils.data import Dataset
from adalflow.utils.file_io import save_csv, save_json, load_json
from adalflow.datasets.utils import prepare_dataset_path
from adalflow.core.base_data_class import DataClass
from adalflow.datasets.types import HotPotQAData


class HotPotQA(Dataset):
    def __init__(
        self,
        only_hard_examples=True,
        root: str = None,
        split: Literal["train", "val", "test"] = "train",
        keep_details: Literal["all", "dev_titles", "none"] = "dev_titles",
        size: int = None,
        **kwargs,
    ) -> None:
        r"""
        official_train: 15661
        sampled_trainset: 11745
        sampled_valset: 3916
        test: 7405

        All answers are a phrase in the supporting context where we can choose supporting facts from the context.

        You can specify the size of the dataset to load by setting the size parameter.
        """
        if split not in ["train", "val", "test"]:
            raise ValueError("Split must be one of 'train', 'val', 'test'")

        if keep_details not in ["all", "dev_titles", "none"]:
            raise ValueError("Keep details must be one of 'all', 'dev_titles', 'none'")

        # if root is None:
        #     root = get_adalflow_default_root_path()
        #     print(f"Saving dataset to {root}")
        self.root = root
        self.task_name = f"hotpot_qa_{keep_details}"
        data_path = prepare_dataset_path(self.root, self.task_name)
        # download and save
        split_csv_path = os.path.join(data_path, f"{split}.json")
        print(f"split_csv_path: {split_csv_path}")
        self._check_or_download_dataset(
            split_csv_path, split, only_hard_examples, keep_details
        )

        # load from csv
        self.data = []
        # created_data_class = DynamicDataClassFactory.from_dict(
        #  "HotPotQAData", {"id": "str", "question": "str", "answer": "str"}

        # with open(split_csv_path, newline="") as csvfile:
        #     reader = csv.DictReader(csvfile)
        #     for i, row in enumerate(reader):
        #         if size is not None and i >= size:
        #             break
        #         self.data.append(HotPotQAData.from_dict(row))

        self.data = load_json(split_csv_path)
        if size is not None:
            # use random seed to make sure the same data is loaded
            # random.Random(0).shuffle(self.data)
            self.data = self.data[:size]
        # convert to dataclass
        self.data = [HotPotQAData.from_dict(d) for d in self.data]

    def _check_or_download_dataset(
        self,
        data_path: str = None,
        split: str = "train",
        only_hard_examples=True,
        keep_details="dev_titles",
    ):
        r"""It will download data from huggingface datasets and split it and save it into three csv files.
        Args:
            data_path (str): The path to save the data. In particular with split name appended.
            split (str): The dataset split, supports ``"train"`` (default), ``"val"`` and ``"test"``. Decides which split to return.
            only_hard_examples (bool): If True, only hard examples will be downloaded.
            keep_details (str): If "all", all details will be kept. If "dev_titles", only dev titles will be kept.
        """

        if data_path is None:
            raise ValueError("data_path must be specified")

        if os.path.exists(data_path):
            return

        safe_import(
            OptionalPackages.DATASETS.value[0], OptionalPackages.DATASETS.value[1]
        )
        from datasets import load_dataset

        assert only_hard_examples, (
            "Care must be taken when adding support for easy examples."
            "Dev must be all hard to match official dev, but training can be flexible."
        )

        hf_official_train = load_dataset(
            "hotpot_qa", "fullwiki", split="train", trust_remote_code=True
        )
        hf_official_dev = load_dataset(
            "hotpot_qa", "fullwiki", split="validation", trust_remote_code=True
        )
        data_path_dir = os.path.dirname(data_path)
        # save all the original data
        all_original_keys = hf_official_train[0].keys()
        for split, examples in zip(
            ["hf_official_train", "hf_official_dev"],
            [hf_official_train, hf_official_dev],
        ):
            target_path = os.path.join(data_path_dir, f"{split}.csv")
            save_csv(examples, f=target_path, fieldnames=all_original_keys)
            # for example in examples:
            #     # is answer in the context
            #     print(f"example: {example}")
            #     context = str(json.dumps(example["context"]))
            #     if example["answer"] in context:
            #         print(f"answer in context")
            #     else:
            #         print(f"answer not in context")
            print(f"saved {split} to {target_path}")
        keys = ["question", "answer"]
        if keep_details == "all":
            keys = [
                "id",
                "question",
                "answer",
                "type",
                "supporting_facts",
                "context",
            ]
        elif keep_details == "dev_titles":
            keys = ["id", "question", "answer", "supporting_facts", "context"]

        official_train = []  # 15661
        for raw_example in hf_official_train:
            if raw_example["level"] == "hard":
                example = {k: raw_example[k] for k in keys}

                if "supporting_facts" in example:
                    example["gold_titles"] = set(example["supporting_facts"]["title"])
                    # del example["supporting_facts"]

                official_train.append(example)
        print(f"official_train: {len(official_train)}")

        rng = random.Random(0)
        rng.shuffle(official_train)

        sampled_trainset = official_train[: len(official_train) * 70 // 100]  # 11745
        print(f"sampled_trainset: {len(sampled_trainset)}")

        sampled_valset = official_train[  # 3916
            len(official_train) * 70 // 100 :
        ]  # this is not the official dev set

        print(f"sampled_valset: {len(sampled_valset)}")

        # for example in self._train:
        #     if keep_details == "dev_titles":
        #         del example["gold_titles"]

        test = []  # 7405

        print(f"raw_example: {hf_official_dev[0]}")
        for raw_example in hf_official_dev:
            assert raw_example["level"] == "hard"
            example = {
                k: raw_example[k]
                for k in ["id", "question", "answer", "type", "supporting_facts"]
            }
            if "supporting_facts" in example:
                example["gold_titles"] = set(example["supporting_facts"]["title"])

                # del example["supporting_facts"]
            test.append(example)

        keys = ["id", "question", "answer", "gold_titles", "context"]

        # split test into val and test
        # random shuff the test
        rng.shuffle(test)
        test_split = test[: len(test) * 50 // 100]  # 3702
        val_split = test[len(test) * 50 // 100 :]  # 3703

        # save to csv
        for split, examples in zip(
            ["train", "val", "test"],
            [sampled_trainset, val_split, test_split],
        ):
            # target_path = prepare_dataset_path(self.root, task_name, split)
            target_path = os.path.join(data_path_dir, f"{split}.json")
            # filter the examples with only the keys
            save_examples: List[HotPotQAData] = []
            for example in examples:
                save_example = {k: example[k] for k in keys if k in example}
                save_example = HotPotQAData.from_dict(save_example)
                save_examples.append(save_example.to_dict())
            save_json(save_examples, f=target_path)
            if split == "train":
                print(f"train example: {examples[0]}")
            print(f"saved {split} to {target_path}")

        if split == "train":
            return sampled_trainset
        elif split == "val":
            return sampled_valset
        elif split == "test":
            return test
        else:
            raise ValueError("Split must be one of 'train', 'val', 'test'")

    def __getitem__(self, index) -> DataClass:
        return self.data[index]

    def __len__(self):
        return len(self.data)


if __name__ == "__main__":
    dataset = HotPotQA(split="train")
    print(dataset[0], type(dataset[0]))
    print(len(dataset))
    valdataset = HotPotQA(split="val")
    print(len(valdataset))
    testdataset = HotPotQA(split="test")
    print(len(testdataset))

    # example = {
    #     "id": "5a8b57f25542995d1e6f1371",
    #     "question": "Were Scott Derrickson and Ed Wood of the same nationality?",
    #     "answer": "yes",
    #     "type": "comparison",
    #     "level": "hard",
    #     "supporting_facts": {
    #         "title": ["Scott Derrickson", "Ed Wood"],
    #         "sent_id": [0, 0],
    #     },
    #     "context": {
    #         "title": [
    #             "Adam Collis",
    #             "Ed Wood (film)",
    #             "Tyler Bates",
    #             "Doctor Strange (2016 film)",
    #             "Hellraiser: Inferno",
    #             "Sinister (film)",
    #             "Deliver Us from Evil (2014 film)",
    #             "Woodson, Arkansas",
    #             "Conrad Brooks",
    #             "The Exorcism of Emily Rose",
    #         ],
    #         "sentences": [
    #             [
    #                 "Adam Collis is an American filmmaker and actor.",
    #                 " He attended the Duke University from 1986 to 1990 and the University of California, Los Angeles from 2007 to 2010.",
    #                 " He also studied cinema at the University of Southern California from 1991 to 1997.",
    #                 ' Collis first work was the assistant director for the Scott Derrickson\'s short "Love in the Ruins" (1995).',
    #                 ' In 1998, he played "Crankshaft" in Eric Koyanagi\'s "Hundred Percent".',
    #             ],
    #             [
    #                 "Ed Wood is a 1994 American biographical period comedy-drama film directed and produced by Tim Burton, and starring Johnny Depp as cult filmmaker Ed Wood.",
    #                 " The film concerns the period in Wood's life when he made his best-known films as well as his relationship with actor Bela Lugosi, played by Martin Landau.",
    #                 " Sarah Jessica Parker, Patricia Arquette, Jeffrey Jones, Lisa Marie, and Bill Murray are among the supporting cast.",
    #             ],
    #             [
    #                 "Tyler Bates (born June 5, 1965) is an American musician, music producer, and composer for films, television, and video games.",
    #                 ' Much of his work is in the action and horror film genres, with films like "Dawn of the Dead, 300, Sucker Punch," and "John Wick."',
    #                 " He has collaborated with directors like Zack Snyder, Rob Zombie, Neil Marshall, William Friedkin, Scott Derrickson, and James Gunn.",
    #                 ' With Gunn, he has scored every one of the director\'s films; including "Guardians of the Galaxy", which became one of the highest grossing domestic movies of 2014, and its 2017 sequel.',
    #                 ' In addition, he is also the lead guitarist of the American rock band Marilyn Manson, and produced its albums "The Pale Emperor" and "Heaven Upside Down".',
    #             ],
    #             [
    #                 "Doctor Strange is a 2016 American superhero film based on the Marvel Comics character of the same name, produced by Marvel Studios and distributed by Walt Disney Studios Motion Pictures.",
    #                 " It is the fourteenth film of the Marvel Cinematic Universe (MCU).",
    #                 " The film was directed by Scott Derrickson, who wrote it with Jon Spaihts and C. Robert Cargill, and stars Benedict Cumberbatch as Stephen Strange, along with Chiwetel Ejiofor, Rachel McAdams, Benedict Wong, Michael Stuhlbarg, Benjamin Bratt, Scott Adkins, Mads Mikkelsen, and Tilda Swinton.",
    #                 ' In "Doctor Strange", surgeon Strange learns the mystic arts after a career-ending car accident.',
    #             ],
    #             [
    #                 "Hellraiser: Inferno (also known as Hellraiser V: Inferno) is a 2000 American horror film.",
    #                 ' It is the fifth installment in the "Hellraiser" series and the first "Hellraiser" film to go straight-to-DVD.',
    #                 " It was directed by Scott Derrickson and released on October 3, 2000.",
    #                 " The film concerns a corrupt detective who discovers Lemarchand's box at a crime scene.",
    #                 " The film's reviews were mixed.",
    #             ],
    #             [
    #                 "Sinister is a 2012 supernatural horror film directed by Scott Derrickson and written by Derrickson and C. Robert Cargill.",
    #                 " It stars Ethan Hawke as fictional true-crime writer Ellison Oswalt who discovers a box of home movies in his attic that puts his family in danger.",
    #             ],
    #             [
    #                 "Deliver Us from Evil is a 2014 American supernatural horror film directed by Scott Derrickson and produced by Jerry Bruckheimer.",
    #                 ' The film is officially based on a 2001 non-fiction book entitled "Beware the Night" by Ralph Sarchie and Lisa Collier Cool, and its marketing campaign highlighted that it was "inspired by actual accounts".',
    #                 " The film stars Eric Bana, Édgar Ramírez, Sean Harris, Olivia Munn, and Joel McHale in the main roles and was released on July 2, 2014.",
    #             ],
    #             [
    #                 "Woodson is a census-designated place (CDP) in Pulaski County, Arkansas, in the United States.",
    #                 " Its population was 403 at the 2010 census.",
    #                 " It is part of the Little Rock–North Little Rock–Conway Metropolitan Statistical Area.",
    #                 " Woodson and its accompanying Woodson Lake and Wood Hollow are the namesake for Ed Wood Sr., a prominent plantation owner, trader, and businessman at the turn of the 20th century.",
    #                 " Woodson is adjacent to the Wood Plantation, the largest of the plantations own by Ed Wood Sr.",
    #             ],
    #             [
    #                 "Conrad Brooks (born Conrad Biedrzycki on January 3, 1931 in Baltimore, Maryland) is an American actor.",
    #                 " He moved to Hollywood, California in 1948 to pursue a career in acting.",
    #                 ' He got his start in movies appearing in Ed Wood films such as "Plan 9 from Outer Space", "Glen or Glenda", and "Jail Bait."',
    #                 " He took a break from acting during the 1960s and 1970s but due to the ongoing interest in the films of Ed Wood, he reemerged in the 1980s and has become a prolific actor.",
    #                 " He also has since gone on to write, produce and direct several films.",
    #             ],
    #             [
    #                 "The Exorcism of Emily Rose is a 2005 American legal drama horror film directed by Scott Derrickson and starring Laura Linney and Tom Wilkinson.",
    #                 " The film is loosely based on the story of Anneliese Michel and follows a self-proclaimed agnostic who acts as defense counsel (Linney) representing a parish priest (Wilkinson), accused by the state of negligent homicide after he performed an exorcism.",
    #             ],
    #         ],
    #     },
    # }

    # # save to csv
    # keys = ["id", "question", "answer", "gold_titles", "context"]
    # example["gold_titles"] = set(example["supporting_facts"]["title"])

    # # test, save to hotpotQA

    # data = HotPotQAData.from_dict({k: example[k] for k in keys})
    # print(f"data: {data}")

    # # save to json
    # save_json([data.to_dict()], f="test.json")

    # # load from json
    # loaded_data = load_json("test.json")
    # # convert to dataclass
    # data = HotPotQAData.from_dict(loaded_data[0])
    # print(f"data: {data}")



"""AWS Bedrock ModelClient integration."""

import json
import os
from typing import Dict, Optional, Any, Callable, Generator as GeneratorType
import backoff
import logging

from adalflow.core.model_client import ModelClient
from adalflow.core.types import ModelType, CompletionUsage, GeneratorOutput

from adalflow.utils.lazy_import import safe_import, OptionalPackages

import sys

boto3_modules = safe_import(
    OptionalPackages.BOTO3.value[0],  # List of package names
    OptionalPackages.BOTO3.value[1],  # Error message
)
# Manually add each module to sys.modules to make them available globally as if imported normally
boto3_module_names = OptionalPackages.BOTO3.value[0]
for name, module in zip(boto3_module_names, boto3_modules):
    sys.modules[name] = module
import boto3
from botocore.config import Config

log = logging.getLogger(__name__)

bedrock_runtime_exceptions = boto3.client(
    service_name="bedrock-runtime",
    region_name=os.getenv("AWS_REGION_NAME", "us-east-1"),
).exceptions


def get_first_message_content(completion: Dict) -> str:
    r"""When we only need the content of the first message.
    It is the default parser for chat completion."""
    return completion["output"]["message"]["content"][0]["text"]


__all__ = [
    "BedrockAPIClient",
    "get_first_message_content",
    "bedrock_runtime_exceptions",
]


class BedrockAPIClient(ModelClient):
    __doc__ = r"""A component wrapper for the Bedrock API client.

    Note:

    This api is in experimental and is not fully tested and validated yet.

    Support:
    1. AWS Titan
    2. Claude
    3. Cohere
    4. LLama
    5. Mistral
    6. Jamba

    Setup:
    1. Install boto3: `pip install boto3`
    2. Ensure you have the AWS credentials set up. There are four variables you can optionally set:
        - AWS_PROFILE_NAME: The name of the AWS profile to use.
        - AWS_REGION_NAME: The name of the AWS region to use.
        - AWS_ACCESS_KEY_ID: The AWS access key ID.
        - AWS_SECRET_ACCESS_KEY: The AWS secret access key.




    Example:

    .. code-block:: python

        from adalflow.components.model_client import BedrockAPIClient

        template = "<SYS>
        You are a helpful assistant.
        </SYS>
        User: {{input_str}}
        You:
        "

        # use AWS_PROFILE_NAME and AWS_REGION_NAME from the environment variables in this case
        # ensure you request the modelId from the AWS team
        self.generator = Generator(
            model_client=BedrockAPIClient(),
            model_kwargs={
                "modelId": "anthropic.claude-3-sonnet-20240229-v1:0",
                "inferenceConfig": {
                    "temperature": 0.8
                }
            }, template=template
        )

    Relevant API docs:
    1. https://docs.aws.amazon.com/bedrock/latest/APIReference/welcome.html
    2. https://docs.aws.amazon.com/bedrock/latest/userguide/getting-started-api-ex-python.html
    3. https://docs.aws.amazon.com/bedrock/latest/APIReference/API_runtime_Converse.html#API_runtime_Converse_RequestParameters
    4. To setup the AWS credentials, follow the instructions here:
    https://docs.aws.amazon.com/bedrock/latest/userguide/getting-started.html
    5. Additionally, this medium article is a good reference:
    https://medium.com/@harangpeter/setting-up-aws-bedrock-for-api-based-text-inference-dc25ab2b216b
    """

    def __init__(
        self,
        aws_profile_name="default",
        aws_region_name="us-west-2",  # Use a supported default region
        aws_access_key_id=None,
        aws_secret_access_key=None,
        aws_session_token=None,
        aws_connection_timeout=None,
        aws_read_timeout=None,
        chat_completion_parser: Callable = None,
    ):
        super().__init__()
        self._aws_profile_name = aws_profile_name
        self._aws_region_name = aws_region_name
        self._aws_access_key_id = aws_access_key_id
        self._aws_secret_access_key = aws_secret_access_key
        self._aws_session_token = aws_session_token
        self._aws_connection_timeout = aws_connection_timeout
        self._aws_read_timeout = aws_read_timeout

        self._client = None
        self.session = None
        self.sync_client = self.init_sync_client()
        self.chat_completion_parser = (
            chat_completion_parser or get_first_message_content
        )

    def init_sync_client(self):
        """
        There is no need to pass both profile and secret key and access key. Path one of them.
        if the compute power assume a role that have access to bedrock, no need to pass anything.
        """
        aws_profile_name = self._aws_profile_name or os.getenv("AWS_PROFILE_NAME")
        aws_region_name = self._aws_region_name or os.getenv("AWS_REGION_NAME")
        aws_access_key_id = self._aws_access_key_id or os.getenv("AWS_ACCESS_KEY_ID")
        aws_secret_access_key = self._aws_secret_access_key or os.getenv(
            "AWS_SECRET_ACCESS_KEY"
        )
        aws_session_token = self._aws_session_token or os.getenv("AWS_SESSION_TOKEN")

        config = None
        if self._aws_connection_timeout or self._aws_read_timeout:
            config = Config(
                connect_timeout=self._aws_connection_timeout,  # Connection timeout in seconds
                read_timeout=self._aws_read_timeout,  # Read timeout in seconds
            )

        session = boto3.Session(
            profile_name=aws_profile_name,
            region_name=aws_region_name,
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
            aws_session_token=aws_session_token,
        )
        bedrock_runtime = session.client(service_name="bedrock-runtime", config=config)

        self._client = session.client(service_name="bedrock")
        return bedrock_runtime

    def init_async_client(self):
        raise NotImplementedError("Async call not implemented yet.")

    def handle_stream_response(self, stream: dict) -> GeneratorType:
        r"""Handle the stream response from bedrock. Yield the chunks.

        Args:
            stream (dict): The stream response generator from bedrock.

        Returns:
            GeneratorType: A generator that yields the chunks from bedrock stream.
        """
        try:
            stream: GeneratorType = stream["stream"]
            for chunk in stream:
                log.debug(f"Raw chunk: {chunk}")
                yield chunk
        except Exception as e:
            log.debug(f"Error in handle_stream_response: {e}")  # Debug print
            raise

    def parse_chat_completion(self, completion: dict) -> "GeneratorOutput":
        r"""Parse the completion, and assign it into the raw_response attribute.

        If the completion is a stream, it will be handled by the handle_stream_response
        method that returns a Generator. Otherwise, the completion will be parsed using
        the get_first_message_content method.

        Args:
            completion (dict): The completion response from bedrock API call.

        Returns:
            GeneratorOutput: A generator output object with the parsed completion. May
                return a generator if the completion is a stream.
        """
        try:
            usage = None
            data = self.chat_completion_parser(completion)
            if not isinstance(data, GeneratorType):
                # Streaming completion usage tracking is not implemented.
                usage = self.track_completion_usage(completion)
            return GeneratorOutput(
                data=None, error=None, raw_response=data, usage=usage
            )
        except Exception as e:
            log.error(f"Error parsing the completion: {e}")
            return GeneratorOutput(
                data=None, error=str(e), raw_response=json.dumps(completion)
            )

    def track_completion_usage(self, completion: Dict) -> CompletionUsage:
        r"""Track the completion usage."""
        usage = completion["usage"]
        return CompletionUsage(
            completion_tokens=usage["outputTokens"],
            prompt_tokens=usage["inputTokens"],
            total_tokens=usage["totalTokens"],
        )

    def list_models(self):
        # Initialize Bedrock client (not runtime)

        try:
            response = self._client.list_foundation_models()
            models = response.get("models", [])
            for model in models:
                print(f"Model ID: {model['modelId']}")
                print(f"  Name: {model['name']}")
                print(f"  Description: {model['description']}")
                print(f"  Provider: {model['provider']}")
                print("")

        except Exception as e:
            print(f"Error listing models: {e}")

    def convert_inputs_to_api_kwargs(
        self,
        input: Optional[Any] = None,
        model_kwargs: Dict = {},
        model_type: ModelType = ModelType.UNDEFINED,
    ):
        """
        check the converse api doc here:
        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/bedrock-runtime/client/converse.html
        """
        api_kwargs = model_kwargs.copy()
        if model_type == ModelType.LLM:
            api_kwargs["messages"] = [
                {"role": "user", "content": [{"text": input}]},
            ]
        else:
            raise ValueError(f"Model type {model_type} not supported")
        return api_kwargs

    @backoff.on_exception(
        backoff.expo,
        (
            bedrock_runtime_exceptions.ThrottlingException,
            bedrock_runtime_exceptions.ModelTimeoutException,
            bedrock_runtime_exceptions.InternalServerException,
            bedrock_runtime_exceptions.ModelErrorException,
            bedrock_runtime_exceptions.ValidationException,
        ),
        max_time=2,
    )
    def call(
        self,
        api_kwargs: Dict = {},
        model_type: ModelType = ModelType.UNDEFINED,
    ) -> dict:
        """
        kwargs is the combined input and model_kwargs
        """
        if model_type == ModelType.LLM:
            if "stream" in api_kwargs and api_kwargs.get("stream", False):
                log.debug("Streaming call")
                api_kwargs.pop(
                    "stream", None
                )  # stream is not a valid parameter for bedrock
                self.chat_completion_parser = self.handle_stream_response
                return self.sync_client.converse_stream(**api_kwargs)
            else:
                api_kwargs.pop("stream", None)
                return self.sync_client.converse(**api_kwargs)
        else:
            raise ValueError(f"model_type {model_type} is not supported")

    async def acall(self):
        raise NotImplementedError("Async call not implemented yet.")



"""Cohere ModelClient integration."""

import os
from typing import Dict, Optional, Any
import backoff
from adalflow.utils.lazy_import import safe_import, OptionalPackages

cohere = safe_import(OptionalPackages.COHERE.value[0], OptionalPackages.COHERE.value[1])
from cohere import (
    BadRequestError,
    InternalServerError,
)


from adalflow.core.model_client import ModelClient
from adalflow.core.types import ModelType


class CohereAPIClient(ModelClient):
    __doc__ = r"""A component wrapper for the Cohere API.

    Visit https://docs.cohere.com/ for more api details.

    References:
    - Cohere reranker: https://docs.cohere.com/reference/rerank

    Tested Cohere models: 6/16/2024
    -  rerank-english-v3.0, rerank-multilingual-v3.0, rerank-english-v2.0, rerank-multilingual-v2.0

    .. note::
        For all ModelClient integration, such as CohereAPIClient, if you want to subclass CohereAPIClient, you need to import it from the module directly.

        ``from adalflow.components.model_client.cohere_client import CohereAPIClient``

        instead of using the lazy import with:

        ``from adalflow.components.model_client import CohereAPIClient``
    """

    def __init__(self, api_key: Optional[str] = None):
        r"""It is recommended to set the GROQ_API_KEY environment variable instead of passing it as an argument.

        Args:
            api_key (Optional[str], optional): Groq API key. Defaults to None.
        """
        super().__init__()
        self._api_key = api_key
        self.init_sync_client()

        self.async_client = None  # only initialize if the async call is called

    def init_sync_client(self):
        api_key = self._api_key or os.getenv("COHERE_API_KEY")
        if not api_key:
            raise ValueError("Environment variable COHERE_API_KEY must be set")
        self.sync_client = cohere.Client(api_key=api_key)

    def init_async_client(self):
        api_key = self._api_key or os.getenv("COHERE_API_KEY")
        if not api_key:
            raise ValueError("Environment variable COHERE_API_KEY must be set")
        self.async_client = cohere.AsyncClient(api_key=api_key)

    def convert_inputs_to_api_kwargs(
        self,
        input: Optional[Any] = None,  # for retriever, it is a list of string.
        model_kwargs: Dict = {},
        model_type: ModelType = ModelType.UNDEFINED,
    ) -> Dict:
        r"""
        For rerank model, expect model_kwargs to have the following keys:
         model: str,
         query: str,
         documents: List[str],
         top_n: int,
        """
        final_model_kwargs = model_kwargs.copy()
        if model_type == ModelType.RERANKER:
            final_model_kwargs["query"] = input
            if "model" not in final_model_kwargs:
                raise ValueError("model must be specified")
            if "documents" not in final_model_kwargs:
                raise ValueError("documents must be specified")
            if "top_k" not in final_model_kwargs:
                raise ValueError("top_k must be specified")

            # convert top_k to the api specific, which is top_n
            final_model_kwargs["top_n"] = final_model_kwargs.pop("top_k")
            return final_model_kwargs
        else:
            raise ValueError(f"model_type {model_type} is not supported")

    @backoff.on_exception(
        backoff.expo,
        (
            BadRequestError,
            InternalServerError,
        ),
        max_time=5,
    )
    def call(self, api_kwargs: Dict = {}, model_type: ModelType = ModelType.UNDEFINED):
        assert (
            "model" in api_kwargs
        ), f"model must be specified in api_kwargs: {api_kwargs}"
        if (
            model_type == ModelType.RERANKER
        ):  # query -> # scores for top_k documents, index for the top_k documents, return as tuple

            response = self.sync_client.rerank(**api_kwargs)
            top_k_scores = [result.relevance_score for result in response.results]
            top_k_indices = [result.index for result in response.results]
            return top_k_indices, top_k_scores
        else:
            raise ValueError(f"model_type {model_type} is not supported")

    @backoff.on_exception(
        backoff.expo,
        (
            BadRequestError,
            InternalServerError,
        ),
        max_time=5,
    )
    async def acall(
        self, api_kwargs: Dict = {}, model_type: ModelType = ModelType.UNDEFINED
    ):
        if self.async_client is None:
            self.init_async_client()
        if "model" not in api_kwargs:
            raise ValueError("model must be specified")
        if model_type == ModelType.RERANKER:
            response = await self.async_client.rerank(**api_kwargs)
            top_k_scores = [result.relevance_score for result in response.results]
            top_k_indices = [result.index for result in response.results]
            return top_k_indices, top_k_scores
        else:
            raise ValueError(f"model_type {model_type} is not supported")



r"""We let users install the required SDKs conditionally for our integrated model providers."""

from adalflow.utils.registry import EntityMapping
from adalflow.utils.lazy_import import (
    LazyImport,
    OptionalPackages,
)

# NOTE: Do not subclass lazy imported classes, it will cause issues with the lazy import mechanism.
# Instead, directly import the class from its specif module and use it.
CohereAPIClient = LazyImport(
    "adalflow.components.model_client.cohere_client.CohereAPIClient",
    OptionalPackages.COHERE,
)
TransformerReranker = LazyImport(
    "adalflow.components.model_client.transformers_client.TransformerReranker",
    OptionalPackages.TRANSFORMERS,
)
TransformerEmbedder = LazyImport(
    "adalflow.components.model_client.transformers_client.TransformerEmbedder",
    OptionalPackages.TRANSFORMERS,
)
TransformerLLM = LazyImport(
    "adalflow.components.model_client.transformers_client.TransformerLLM",
    OptionalPackages.TRANSFORMERS,
)
TransformersClient = LazyImport(
    "adalflow.components.model_client.transformers_client.TransformersClient",
    OptionalPackages.TRANSFORMERS,
)
AnthropicAPIClient = LazyImport(
    "adalflow.components.model_client.anthropic_client.AnthropicAPIClient",
    OptionalPackages.ANTHROPIC,
)
BedrockAPIClient = LazyImport(
    "adalflow.components.model_client.bedrock_client.BedrockAPIClient",
    OptionalPackages.BOTO3,
)
GroqAPIClient = LazyImport(
    "adalflow.components.model_client.groq_client.GroqAPIClient",
    OptionalPackages.GROQ,
)
OpenAIClient = LazyImport(
    "adalflow.components.model_client.openai_client.OpenAIClient",
    OptionalPackages.OPENAI,
)
GoogleGenAIClient = LazyImport(
    "adalflow.components.model_client.google_client.GoogleGenAIClient",
    OptionalPackages.GOOGLE_GENERATIVEAI,
)
OllamaClient = LazyImport(
    "adalflow.components.model_client.ollama_client.OllamaClient",
    OptionalPackages.OLLAMA,
)
get_first_message_content = LazyImport(
    "adalflow.components.model_client.openai_client.get_first_message_content",
    OptionalPackages.OPENAI,
)
get_all_messages_content = LazyImport(
    "adalflow.components.model_client.openai_client.get_all_messages_content",
    OptionalPackages.OPENAI,
)
get_probabilities = LazyImport(
    "adalflow.components.model_client.openai_client.get_probabilities",
    OptionalPackages.OPENAI,
)

__all__ = [
    "CohereAPIClient",
    "TransformerReranker",
    "TransformerEmbedder",
    "TransformerLLM",
    "TransformersClient",
    "AnthropicAPIClient",
    "BedrockAPIClient",
    "GroqAPIClient",
    "OpenAIClient",
    "GoogleGenAIClient",
]

for name in __all__:
    EntityMapping.register(name, globals()[name])



"Helpers for model client for integrating models and parsing the output."
from adalflow.core.types import EmbedderOutput, Embedding, Usage


def parse_embedding_response(
    api_response,
) -> EmbedderOutput:
    r"""Parse embedding model output from the API response to EmbedderOutput.

    Follows the OpenAI API response pattern.
    """
    # Assuming `api_response` has `.embeddings` and `.usage` attributes
    # and that `embeddings` is a list of objects that can be converted to `Embedding` dataclass
    # TODO: check if any embedding is missing
    embeddings = [
        Embedding(embedding=e.embedding, index=e.index) for e in api_response.data
    ]
    usage = Usage(
        prompt_tokens=api_response.usage.prompt_tokens,
        total_tokens=api_response.usage.total_tokens,
    )  # Assuming `usage` is an object with a `count` attribute

    # Assuming the model name is part of the response or set statically here
    model = api_response.model

    return EmbedderOutput(data=embeddings, model=model, usage=usage)



"""Extract and convert common string to Python objects.

From simple data types like boolean, integer, and float to more complex data types like JSON, YAML, and list strings."""

from typing import Dict, List, Union
import logging
from adalflow.core.component import DataComponent


import adalflow.core.functional as F

log = logging.getLogger(__name__)

BOOLEAN_PARSER_OUTPUT_TYPE = bool


# class Parser:
#     __doc__ = r"""Base class for all string parsers."""

#     def __init__(self):
#         super().__init__()

#     def __call__(self, input: str) -> object:
#         return self.call(input)

#     def call(self, input: str) -> object:
#         raise NotImplementedError(
#             "Parser subclasses must implement the __call__ method"
#         )


class BooleanParser(DataComponent):
    __doc__ = r"""Extracts boolean values from text.

    Examples:

    .. code-block:: python

        boolean_parser = BooleanParser()
        test_input_1 = "True" # or "true" or "...true..."
        print(boolean_parser(test_input_1))  # Expected to extract True
    """

    def __init__(self, *args, **kwargs):
        super().__init__()

    def call(self, input: str) -> BOOLEAN_PARSER_OUTPUT_TYPE:
        input = input.strip()
        try:
            return F.extract_first_boolean(input)
        except Exception as e:
            raise ValueError(f"Error: {e}")


INT_PARSER_OUTPUT_TYPE = int


class IntParser(DataComponent):
    __doc__ = r"""Extracts integer values from text.

    Returns:
        int: Extracted integer value.

    Raises:
        ValueError: If the input text does not contain an integer

    Examples:

    .. code-block:: python

        int_parser = IntParser()
        test_input_2 = "123" # or "...123..."
        print(int_parser(test_input_2))  # Expected to extract 123
    """

    def __init__(self):
        super().__init__()

    def call(self, input: str) -> INT_PARSER_OUTPUT_TYPE:
        input = input.strip()
        try:
            return F.extract_first_int(input)
        except Exception as e:
            raise ValueError(f"Error: {e}")


FLOAT_PARSER_OUTPUT_TYPE = float


class FloatParser(DataComponent):
    __doc__ = r"""Extracts float values from text.

    Returns:
        float: Extracted float value.

    Raises:
        ValueError: If the input text does not contain a float

    Examples:

    .. code-block:: python

        float_parser = FloatParser()
        test_input_3 = "123.45" # or "...123.45..."
        print(float_parser(test_input_3))  # Expected to extract 123.45
    """

    def __init__(self):
        super().__init__()

    def call(self, input: str) -> FLOAT_PARSER_OUTPUT_TYPE:
        input = input.strip()
        try:
            return F.extract_first_float(input)
        except Exception as e:
            raise ValueError(f"Error: {e}")


LIST_PARSER_OUTPUT_TYPE = List[object]


class ListParser(DataComponent):
    __doc__ = r"""Extracts list `[...]` strings from text and parses them into a list object.

    Args:
        add_missing_right_bracket (bool, optional): Add a missing right bracket to the list string. Defaults to True.

    Returns:
        List[object]: Extracted list object.

    Raises:
        ValueError: If the input text does not contain a list

    Examples:

    .. code-block:: python

        list_parser = ListParser()
        test_input_4 = 'Some random text before ["item1", "item2"] and more after'
        print(list_parser(test_input_4))  # Expected to extract ["item1", "item2"]
    """

    def __init__(self, add_missing_right_bracket: bool = True):
        super().__init__()
        self.add_missing_right_bracket = add_missing_right_bracket

    def call(self, input: str) -> LIST_PARSER_OUTPUT_TYPE:
        input = input.strip()
        list_str = None
        # Extract list string
        try:
            list_str = F.extract_list_str(input, self.add_missing_right_bracket)

        except Exception as e:
            raise ValueError(f"Error at extracting list string: {e}")

        # Parse list string with json.loads and yaml.safe_load
        try:
            list_obj = F.parse_json_str_to_obj(list_str)
            return list_obj
        except Exception as e:
            log.error(f"Error at parsing list string with json.loads: {e}")
            raise ValueError(f"Error: {e}")


JSON_PARSER_OUTPUT_TYPE = Union[Dict[str, object], List[object]]


class JsonParser(DataComponent):
    __doc__ = r"""Extracts JSON strings `{...}` or `[...]` from text and parses them into a JSON object.

    It can output either a dictionary or a list as they are both valid JSON objects.

    Args:
        add_missing_right_brace (bool, optional): Add a missing right brace to the JSON string. Defaults to True.

    Returns:
        Union[Dict[str, object], List[object]]: Extracted JSON object.

    Raises:
        ValueError: If the input text does not contain a JSON object


    Examples:

    .. code-block:: python

        json_parser = JsonParser()
        json_str = "```json\n{\"key\": \"value\"}\n```"
        json_obj = json_parser(json_str)
        print(json_obj)  # Expected to extract {"key": "value"}
    """

    def __init__(self, add_missing_right_brace: bool = True):
        super().__init__()
        self.add_missing_right_brace = add_missing_right_brace

    def call(self, input: str) -> JSON_PARSER_OUTPUT_TYPE:
        input = input.strip()
        # Extract JSON string
        json_str = None
        try:
            json_str = F.extract_json_str(input, self.add_missing_right_brace)
            log.debug(f"json_str: {json_str}")

        except Exception as e:
            raise ValueError(f"Error: {e}")
        # Parse JSON string with json.loads and yaml.safe_load
        try:
            json_obj = F.parse_json_str_to_obj(json_str)
            log.debug(f"json_obj: {json_obj}")
            return json_obj
        except Exception as e:
            log.error(f"Error at parsing JSON string: {e}")
            raise ValueError(f"Error: {e}")


YAML_PARSER_OUTPUT_TYPE = JSON_PARSER_OUTPUT_TYPE


# TODO: yaml parser needs to be more robust, currently json works way better than yaml
class YamlParser(DataComponent):
    __doc__ = r"""To extract YAML strings from text and parse them into a YAML object.

    Returns:
        JSON_PARSER_OUTPUT_TYPE: Extracted YAML object.

    Raises:
        ValueError: If the input text does not contain a YAML object

    Examples:

    .. code-block:: python

        yaml_parser = YamlParser()
        yaml_str = "```yaml\nkey: value\n```"
        yaml_obj = yaml_parser(yaml_str)
        print(yaml_obj)  # Expected to extract {"key": "value"}
    """

    def __init__(self):
        super().__init__()

    def call(self, input: str) -> YAML_PARSER_OUTPUT_TYPE:
        input = input.strip()
        # parse YAML string with yaml.safe_load
        try:
            yaml_str = F.extract_yaml_str(input)
            yaml_obj = F.parse_yaml_str_to_obj(yaml_str)
            return yaml_obj
        except Exception as e:
            raise ValueError(f"Error: {e}")



"""
Tokenizer from tiktoken.
"""

import tiktoken
from typing import List

# from adalflow.core.component import BaseComponent


class Tokenizer:
    __doc__ = r"""
    Tokenizer component that wraps around the tokenizer from tiktoken.
    __call__ is the same as forward/encode, so that we can use it in Sequential
    Additonally, you can can also use encode and decode methods.

    Args:
        name (str, optional): The name of the tokenizer. Defaults to "cl100k_base". You can find more information
        at the tiktoken documentation.
    """

    def __init__(self, name: str = "cl100k_base", remove_stop_words: bool = False):
        super().__init__()
        self.name = name
        self.tokenizer = tiktoken.get_encoding(name)
        self.stop_words = (
            set(["and", "the", "is", "in", "at", "of", "a", "an"])
            if remove_stop_words
            else set()
        )

    # call is the same as forward/encode, so that we can use it in Sequential
    def __call__(self, input: str) -> List[str]:
        return self.encode(input)

    def preprocess(self, text: str) -> List[str]:
        # Lowercase the text
        words = text.lower().split()
        return words

    def encode(self, text: str) -> List[int]:
        r"""Encodes the input text/word into token IDs."""
        return self.tokenizer.encode(text)

    def decode(self, tokens: List[str]) -> str:
        r"""Decodes the input tokens into text."""
        return self.tokenizer.decode(tokens)

    def count_tokens(self, text: str) -> int:
        r"""Counts the number of tokens in the input text."""
        return len(self.encode(text))

    def get_string_tokens(self, text: str) -> List[str]:
        r"""Returns the string tokens from the input text."""
        token_ids = self.encode(text)
        return [self.tokenizer.decode([token_id]) for token_id in token_ids]



"""Ollama ModelClient integration."""

import os
from typing import (
    Dict,
    Optional,
    Any,
    TypeVar,
    List,
    Type,
    Generator as GeneratorType,
    Union,
)
import backoff
import logging
import warnings
from adalflow.core.types import ModelType, GeneratorOutput

from adalflow.utils.lazy_import import safe_import, OptionalPackages

ollama = safe_import(OptionalPackages.OLLAMA.value[0], OptionalPackages.OLLAMA.value[1])
import ollama
from ollama import RequestError, ResponseError, GenerateResponse


from adalflow.core.model_client import ModelClient
from adalflow.core.types import EmbedderOutput, Embedding

log = logging.getLogger(__name__)

T = TypeVar("T")


def parse_stream_response(completion: GeneratorType) -> Any:
    """Parse the completion to a str. We use the generate with prompt instead of chat with messages."""
    for chunk in completion:
        log.debug(f"Raw chunk: {chunk}")
        raw_response = chunk["response"] if "response" in chunk else None
        yield GeneratorOutput(data=None, raw_response=raw_response)


def parse_generate_response(completion: GenerateResponse) -> "GeneratorOutput":
    """Parse the completion to a str. We use the generate with prompt instead of chat with messages."""
    if "response" in completion:
        log.debug(f"response: {completion}")
        raw_response = completion["response"]
        return GeneratorOutput(data=None, raw_response=raw_response)
    else:
        log.error(
            f"Error parsing the completion: {completion}, type: {type(completion)}"
        )
        # raise ValueError(
        #     f"Error parsing the completion: {completion}, type: {type(completion)}"
        # )
        return GeneratorOutput(
            data=None, error="Error parsing the completion", raw_response=completion
        )


class OllamaClient(ModelClient):
    __doc__ = r"""A component wrapper for the Ollama SDK client.

    To make a model work, you need to:

    - [Download Ollama app] Go to https://github.com/ollama/ollama?tab=readme-ov-file to download the Ollama app (command line tool).
      Choose the appropriate version for your operating system.

    - [Pull a model] Run the following command to pull a model:

    .. code-block:: shell

        ollama pull llama3

    - [Run a model] Run the following command to run a model:

    .. code-block:: shell

        ollama run llama3

    This model will be available at http://localhost:11434. You can also chat with the model at the terminal after running the command.

    Args:
        host (Optional[str], optional): Optional host URI.
            If not provided, it will look for OLLAMA_HOST env variable. Defaults to None.
            The default host is "http://localhost:11434".

    Setting model_kwargs:

        For LLM, expect model_kwargs to have the following keys:

        model (str, required):
            Use `ollama list` via your CLI or  visit ollama model page on https://ollama.com/library

        stream (bool, default: False ) – Whether to stream the results.


        options (Optional[dict], optional)
            Options that affect model output.

            # If not specified the following defaults will be assigned.

                "seed": 0, - Sets the random number seed to use for generation. Setting this to a specific number will make the model generate the same text for the same prompt.

                "num_predict": 128, - Maximum number of tokens to predict when generating text. (-1  = infinite generation, -2 = fill context)

                "top_k": 40, - Reduces the probability of generating nonsense. A higher value (e.g. 100) will give more diverse answers, while a lower value (e.g. 10) will be more conservative.

                "top_p": 0.9, - 	Works together with top-k. A higher value (e.g., 0.95) will lead to more diverse text, while a lower value (e.g., 0.5) will generate more focused and conservative text.

                "tfs_z": 1, - Tail free sampling. This is used to reduce the impact of less probable tokens from the output. Disabled by default (e.g. 1) (More documentation here for specifics)

                "repeat_last_n": 64, - Sets how far back the model should look back to prevent repetition. (0 = disabled, -1 = num_ctx)

                "temperature": 0.8, - The temperature of the model. Increasing the temperature will make the model answer more creatively.

                "repeat_penalty": 1.1, - Sets how strongly to penalize repetitions. A higher value(e.g., 1.5 will penlaize repetitions more strongly, while lowe values *e.g., 0.9 will be more lenient.)

                "mirostat": 0.0, - Enable microstat smapling for controlling perplexity. (0 = disabled, 1 = microstat, 2 = microstat 2.0)

                "mirostat_tau": 0.5, - Controls the balance between coherence and diversity of the output. A lower value will result in more focused and coherent text.

                "mirostat_eta": 0.1, - Influences how quickly the algorithm responds to feedback from the generated text. A lower learning rate will result in slower adjustments, while a higher learning rate will make the algorithm more responsive.

                "stop": ["\n", "user:"], - 	Sets the stop sequences to use. When this pattern is encountered the LLM will stop generating text and return. Multiple stop patterns may be set by specifying multiple separate stop parameters in a modelfile.

                "num_ctx": 2048, - Sets the size of the context window used to generate the next token.

        For EMBEDDER, expect model_kwargs to have the following keys:

        model (str, required):
            Use `ollama list` via your CLI or  visit ollama model page on https://ollama.com/library

        prompt (str, required):
            String that is sent to the Embedding model.

        options (Optional[dict], optional):
            See LLM args for defaults.

    References:

        - https://github.com/ollama/ollama-python
        - https://github.com/ollama/ollama
        - Models: https://ollama.com/library
        - Ollama API: https://github.com/ollama/ollama/blob/main/docs/api.md
        - Options Parameters: https://github.com/ollama/ollama/blob/main/docs/modelfile.md.
        - LlamaCPP API documentation(Ollama is based on this): https://llama-cpp-python.readthedocs.io/en/latest/api-reference/#low-level-api
        - LLM API: https://llama-cpp-python.readthedocs.io/en/stable/api-reference/#llama_cpp.Llama.create_completion

    Tested Ollama models: 7/9/24

    -  internlm2:latest
    -  llama3
    -  jina/jina-embeddings-v2-base-en:latest

    .. note::
       We use `embeddings` and `generate` apis from Ollama SDK.
       Please refer to https://github.com/ollama/ollama-python/blob/main/ollama/_client.py for model_kwargs details.
    """

    def __init__(self, host: Optional[str] = None):
        super().__init__()

        self._host = host or os.getenv("OLLAMA_HOST")
        if not self._host:
            warnings.warn(
                "Better to provide host or set OLLAMA_HOST env variable. We will use the default host http://localhost:11434 for now."
            )
            self._host = "http://localhost:11434"

        log.debug(f"Using host: {self._host}")

        self.init_sync_client()
        self.async_client = None  # only initialize if the async call is called

    def init_sync_client(self):
        """Create the synchronous client"""

        self.sync_client = ollama.Client(host=self._host)

    def init_async_client(self):
        """Create the asynchronous client"""

        self.async_client = ollama.AsyncClient(host=self._host)

    # NOTE: do not put yield and return in the same function, thus we separate the functions
    def parse_chat_completion(
        self, completion: Union[GenerateResponse, GeneratorType]
    ) -> "GeneratorOutput":
        """Parse the completion to a str. We use the generate with prompt instead of chat with messages."""
        log.debug(f"completion: {completion}, {isinstance(completion, GeneratorType)}")
        if isinstance(completion, GeneratorType):  # streaming
            return parse_stream_response(completion)
        else:
            return parse_generate_response(completion)

    def parse_embedding_response(
        self, response: Dict[str, List[float]]
    ) -> EmbedderOutput:
        r"""Parse the embedding response to a structure LightRAG components can understand.
        Pull the embedding from response['embedding'] and store it Embedding dataclass
        """
        try:
            embeddings = Embedding(embedding=response["embedding"], index=0)
            return EmbedderOutput(data=[embeddings])
        except Exception as e:
            log.error(f"Error parsing the embedding response: {e}")
            return EmbedderOutput(data=[], error=str(e), raw_response=response)

    def convert_inputs_to_api_kwargs(
        self,
        input: Optional[Any] = None,
        model_kwargs: Dict = {},
        model_type: ModelType = ModelType.UNDEFINED,
    ) -> Dict:
        r"""Convert the input and model_kwargs to api_kwargs for the Ollama SDK client."""
        # TODO: ollama will support batch embedding in the future: https://ollama.com/blog/embedding-models
        final_model_kwargs = model_kwargs.copy()
        if model_type == ModelType.EMBEDDER:
            if isinstance(input, str):
                final_model_kwargs["prompt"] = input
                return final_model_kwargs
            else:
                raise ValueError(
                    "Ollama does not support batch embedding yet. It only accepts a single string for input for now. Make sure you are not passing a list of strings"
                )
        elif model_type == ModelType.LLM:
            if input is not None and input != "":
                final_model_kwargs["prompt"] = input
                return final_model_kwargs
            else:
                raise ValueError("Input must be text")
        else:
            raise ValueError(f"model_type {model_type} is not supported")

    @backoff.on_exception(
        backoff.expo,
        (RequestError, ResponseError),
        max_time=5,
    )
    def call(self, api_kwargs: Dict = {}, model_type: ModelType = ModelType.UNDEFINED):
        if "model" not in api_kwargs:
            raise ValueError("model must be specified")
        log.info(f"api_kwargs: {api_kwargs}")
        if not self.sync_client:
            self.init_sync_client()
        if self.sync_client is None:
            raise RuntimeError("Sync client is not initialized")
        if model_type == ModelType.EMBEDDER:
            return self.sync_client.embeddings(**api_kwargs)
        if model_type == ModelType.LLM:
            return self.sync_client.generate(**api_kwargs)
        else:
            raise ValueError(f"model_type {model_type} is not supported")

    @backoff.on_exception(
        backoff.expo,
        (RequestError, ResponseError),
        max_time=5,
    )
    async def acall(
        self, api_kwargs: Dict = {}, model_type: ModelType = ModelType.UNDEFINED
    ):
        if self.async_client is None:
            self.init_async_client()
        if self.async_client is None:
            raise RuntimeError("Async client is not initialized")
        if "model" not in api_kwargs:
            raise ValueError("model must be specified")
        if model_type == ModelType.EMBEDDER:
            return await self.async_client.embeddings(**api_kwargs)
        if model_type == ModelType.LLM:
            return await self.async_client.generate(**api_kwargs)
        else:
            raise ValueError(f"model_type {model_type} is not supported")

    @classmethod
    def from_dict(cls: Type["OllamaClient"], data: Dict[str, Any]) -> "OllamaClient":
        obj = super().from_dict(data)
        # recreate the existing clients
        obj.sync_client = obj.init_sync_client()
        obj.async_client = obj.init_async_client()
        return obj

    def to_dict(self, exclude: Optional[List[str]] = None) -> Dict[str, Any]:
        r"""Convert the component to a dictionary."""

        # combine the exclude list
        exclude = list(set(exclude or []) | {"sync_client", "async_client"})

        output = super().to_dict(exclude=exclude)
        return output


# TODO: add tests to stream and non stream case
# if __name__ == "__main__":
#     from adalflow.core.generator import Generator
#     from adalflow.components.model_client import OllamaClient, OpenAIClient
#     from adalflow.utils import setup_env, get_logger

#     log = get_logger(level="DEBUG")

#     setup_env()

#     ollama_ai = {
#         "model_client": OllamaClient(),
#         "model_kwargs": {
#             "model": "qwen2:0.5b",
#             "stream": True,
#         },
#     }
#     open_ai = {
#         "model_client": OpenAIClient(),
#         "model_kwargs": {
#             "model": "gpt-3.5-turbo",
#             "stream": False,
#         },
#     }
#     # generator = Generator(**open_ai)
#     # output = generator({"input_str": "What is the capital of France?"})
#     # print(output)

#     # generator = Generator(**ollama_ai)
#     # output = generator({"input_str": "What is the capital of France?"})
#     # print(output)



"""Adapted from text_grad's String Based Function"""

from typing import Callable, Dict, Union, TYPE_CHECKING, Optional
import logging
import json
from adalflow.optim.loss_component import LossComponent
from adalflow.optim.function import BackwardContext


if TYPE_CHECKING:
    from adalflow.core import ModelClient
    from adalflow.core.generator import BackwardEngine
from adalflow.core.types import GeneratorOutput
from adalflow.optim.parameter import (
    Parameter,
    OutputParameter,
)
from adalflow.optim.gradient import GradientContext, Gradient
from adalflow.optim.types import ParameterType

from adalflow.core.prompt_builder import Prompt
from adalflow.eval.base import BaseEvaluator
from adalflow.optim.text_grad.backend_engine_prompt import (
    LOSS_CONVERSATION_TEMPLATE_STRING,
    LOSS_CONVERSATION_START_INSTRUCTION_STRING_FN,
    OBJECTIVE_INSTRUCTION_BASE,
    OBJECTIVE_INSTRUCTION_CHAIN,  # often not used
)
from adalflow.utils import printc


log = logging.getLogger(__name__)


class EvalFnToTextLoss(LossComponent):
    __doc__ = """Convert an evaluation function to a text loss.

    LossComponent will take an eval function and output a score
    (usually a float in range [0, 1], and the higher the better, unlike the loss function in model training).

    In math:

    score/loss = eval_fn(y_pred, y_gt)

    The gradident/feedback = d(score)/d(y_pred) will be computed using a backward engine.
    Gradient_context = GradientContext(
        context=conversation_str,
        response_desc=response.role_desc,
        variable_desc=role_desc,
    )

    Args:
        eval_fn: The evaluation function that takes a pair of y and y_gt and returns a score.
        eval_fn_desc: Description of the evaluation function.
        backward_engine: The backward engine to use for the text prompt optimization.
        model_client: The model client to use for the backward engine if backward_engine is not provided.
        model_kwargs: The model kwargs to use for the backward engine if backward_engine is not provided.

    """

    def __init__(
        self,
        eval_fn: Union[Callable, BaseEvaluator],
        eval_fn_desc: str,
        backward_engine: Optional["BackwardEngine"] = None,
        model_client: "ModelClient" = None,
        model_kwargs: Dict[str, object] = None,
    ):
        from adalflow.core.generator import BackwardEngine

        super().__init__()
        self.eval_fn = eval_fn
        self.eval_fn_desc = eval_fn_desc
        self.name = f"{self.__class__.__name__}"

        self.backward_engine = None

        if backward_engine is None:
            log.info(
                "EvalFnToTextLoss: No backward engine provided. Creating one using model_client and model_kwargs."
            )
            if model_client and model_kwargs:

                self.set_backward_engine(backward_engine, model_client, model_kwargs)
        else:
            if not isinstance(backward_engine, BackwardEngine):
                raise TypeError(
                    "EvalFnToTextLoss: backward_engine must be an instance of BackwardEngine."
                )
            self.backward_engine = backward_engine

    def forward(
        self,
        kwargs: Dict[str, Parameter],
        response_desc: str = None,
        metadata: Dict[str, str] = None,  # additional notes on the input kwargs
        id: str = None,
        gt: object = None,
        input: Dict[str, object] = None,
    ) -> Parameter:
        r"""
        Args:
            kwargs: The inputs to the eval_fn.
            response_desc: Description of the output.
            metadata: Additional notes on the input kwargs.
            id: The unique identifier for the data point.
            gt: The ground truth for the evaluation function.
        """
        if response_desc is None:
            response_desc = "Output of EvalFnToTextLoss."

        # validate the type of kwargs
        predesessors = []
        for k, v in kwargs.items():
            if not isinstance(v, Parameter):
                raise TypeError(
                    f"EvalFnToTextLoss: All inputs must be Parameters. Got {type(v)} for {k}."
                )
            if isinstance(v, Parameter):
                predesessors.append(v)
        eval_inputs = {}
        for k, v in kwargs.items():
            eval_inputs[k] = v.eval_input
        score: float = self.eval_fn(**eval_inputs)

        eval_param: Parameter = OutputParameter(
            name=self.name + "_output",
            data=score,
            requires_opt=True,
            role_desc=response_desc,
            score=score,
            param_type=ParameterType.LOSS_OUTPUT,
            data_id=id,
        )
        eval_param.set_gt(gt)
        eval_param.set_predecessors(predesessors)
        eval_param.trace_forward_pass(
            input_args=kwargs,
            full_response=score,
            id=self.id,
            name=self.name,
        )

        log.info(f"EvalFnToTextLoss: Input: {kwargs}, Output: {eval_param}")
        # extract ground truth from eval_inputs, anything
        eval_param.set_grad_fn(
            BackwardContext(
                backward_fn=self.backward,
                backward_engine=self.backward_engine,
                response=eval_param,
                eval_fn_desc=self.eval_fn_desc,
                kwargs=kwargs,
                metadata=metadata,
                ground_truth=gt,
                input=input,
                disable_backward_engine=self._disable_backward_engine,
            )
        )
        return eval_param

    def set_backward_engine(
        self,
        backward_engine: "BackwardEngine" = None,
        model_client: "ModelClient" = None,
        model_kwargs: Dict[str, object] = None,
    ):
        from adalflow.core.generator import BackwardEngine

        self.backward_engine = backward_engine
        if not backward_engine:
            log.info(
                "EvalFnToTextLoss: No backward engine provided. Creating one using model_client and model_kwargs."
            )
            self.backward_engine = BackwardEngine(model_client, model_kwargs)
        else:
            if type(backward_engine) is not BackwardEngine:
                raise TypeError(
                    f"EvalFnToTextLoss: backward_engine must be an instance of BackwardEngine. Got {type(backward_engine)}."
                )

    @staticmethod
    def _backward_through_one_predecessor(
        pred: Parameter,
        kwargs: Dict[str, Parameter],
        response: Parameter,
        eval_fn_desc: str,
        backward_engine: "BackwardEngine",
        ground_truth: object = None,
        is_intermediate_node: bool = False,  # if the node is an intermediate node in the backpropagation chain
        metadata: Dict[str, str] = None,
        input: Dict[str, object] = None,  # system input
        disable_backward_engine: bool = False,
    ):
        if not pred.requires_opt:
            if response.score is not None:
                pred.set_score(response.score)
            log.debug(
                f"EvalFnToTextLoss: Skipping {pred} as it does not require optimization."
            )
            return
        log.debug(
            f"EvalFnToTextLoss: Backward through {pred}, is_intermediate_node: {is_intermediate_node}"
        )

        if pred.check_if_already_computed_gradient_respect_to(response.id):
            log.info(
                f"EvalFnToTextLoss: Gradient already computed for {pred.role_desc} with respect to {response.role_desc}"
            )

            return

        if backward_engine is None:
            log.error(
                "EvalFnToTextLoss: backward_engine is required for text prompt optimization."
            )
            raise ValueError(
                "EvalFnToTextLoss: backward_engine is required for text prompt optimization."
            )

        instruction_str, objective_str = None, None

        # convert kwargs to key, (value, type(eval_input))

        inputs = {}
        for k, v in kwargs.items():
            inputs[k] = (v.get_param_info(), str(type(v.eval_input)))

        # response information
        conversation_str = Prompt(
            LOSS_CONVERSATION_TEMPLATE_STRING,
            prompt_kwargs={
                "system_question": input,
                "inputs": inputs,
                "eval_fn_desc": eval_fn_desc,
                "response_value": response.get_prompt_data(),
                "metadata": json.dumps(metadata) if metadata else None,
            },
        )()

        conv_ins_template = LOSS_CONVERSATION_START_INSTRUCTION_STRING_FN
        obj_ins_template = OBJECTIVE_INSTRUCTION_BASE

        if is_intermediate_node:
            obj_ins_template = OBJECTIVE_INSTRUCTION_CHAIN

        instruction_str = Prompt(
            conv_ins_template,
            prompt_kwargs={
                "variable": pred.get_param_info(),
                "conversation_str": conversation_str,
            },
        )()
        objective_str = Prompt(
            obj_ins_template,
            prompt_kwargs={
                "response_name": response.name,
                "response_desc": response.role_desc,
                "response_gradient": response.data,
            },
        )()

        log.info(f"EvalFnToTextLoss: Instruction: {instruction_str}")
        log.info(f"EvalFnToTextLoss: Objective: {objective_str}")
        log.info(f"EvalFnToTextLoss: Conversation: {conversation_str}")

        # Compute the gradient
        backward_engine_prompt_kwargs = {
            "conversation_sec": instruction_str,
            "objective_instruction_sec": objective_str,
            # "evaluate_variable_instruction_sec": eval_str,
        }
        gradient_value_data = None
        gradient_prompt = None
        if not disable_backward_engine:
            gradient_value: GeneratorOutput = backward_engine(
                prompt_kwargs=backward_engine_prompt_kwargs
            )
            gradient_prompt = backward_engine.get_prompt(
                **backward_engine_prompt_kwargs
            )
            # print(f"Backward engine prompt: {gradient_prompt}")
            gradient_value_data = (
                gradient_value.data
                or backward_engine.failure_message_to_optimizer(
                    gradient_response=gradient_value
                )
            )

            gradient_value_data = (
                f"expected answer: {ground_truth},\n Feedback: {gradient_value_data}"
            )
            # print(f"gradient_value_data: {gradient_value_data}")

            log.debug(f"EvalFnToTextLoss: Gradient for {pred}: {gradient_value_data}")

        # score should be passed to grad
        gradient_param = Gradient(
            data=gradient_value_data,
            data_id=response.data_id,
            score=response.data,
            from_response=response,
            to_pred=pred,
        )
        gradient_param.add_prompt(gradient_prompt)
        gradient_param.add_context(
            GradientContext(
                input_output=conversation_str,
                response_desc=response.role_desc,
                variable_desc=pred.role_desc,
                # input=input,
                # ground_truth=ground_truth,
            )
        )
        pred.add_gradient(gradient_param)

        # backward the end to end score
        # TODO: not really useful
        if response.score is not None:
            pred.set_score(response.score)
        pred.set_gt(ground_truth)
        printc(f"pred: {pred.eval_input}, gt: {ground_truth}")
        # print(f"setting pred name {pred.name} score to {response.data}")
        # print(f"gradient_param: {pred.gradients}")

        # TODO: reduce meta

    def backward(
        self,
        response: Parameter,
        eval_fn_desc: str,
        kwargs: Dict[str, Parameter],
        ground_truth: object = None,
        backward_engine: Optional[
            "BackwardEngine"
        ] = None,  # only needed for text prompt optimization
        metadata: Dict[str, str] = None,
        input: Dict[str, object] = None,
        disable_backward_engine: bool = False,
    ):
        r"""Ensure to set backward_engine for the text prompt optimization. It can be None if you
        are only doing demo optimization and it will not have gradients but simply backpropagate the score.
        """
        log.info(f"EvalFnToTextLoss: Backward: {response}")
        children_params = response.predecessors
        is_intermediate_node = False
        response_gradient_context = response.get_gradient_and_context_text().strip()
        if response_gradient_context != "":
            log.info("EvalFnToTextLoss is an intermediate node.")
            is_intermediate_node = True
        log.info(f"response_gradient_context: {response_gradient_context}")

        # go through all child parameters
        if backward_engine:
            for pred in children_params:
                if not pred.requires_opt:
                    log.debug(
                        f"EvalFnToTextLoss: Skipping {pred} as it does not require optimization."
                    )
                    continue

                self._backward_through_one_predecessor(
                    pred,
                    kwargs,
                    response,
                    eval_fn_desc,
                    backward_engine,
                    ground_truth=ground_truth,
                    is_intermediate_node=is_intermediate_node,
                    metadata=metadata,
                    input=input,
                    disable_backward_engine=disable_backward_engine,
                )
            # else:  # recursively disable backward for all children
            #     for pred in children_params:
            #         pred.backward_engine_disabled = True
        # backward for the score for the demo
        for pred in children_params:

            if not (isinstance(response.data, float) or isinstance(response.data, int)):
                raise TypeError(
                    f"EvalFnToTextLoss: response.data must be a float. Got {type(response.data)}."
                )
            pred.score = response.data
            from adalflow.utils.logger import printc

            printc(
                f"EvalFnToTextLoss: {pred.name} set_score: {response.data}, {response.name}",
                "blue",
            )
            log.info(f"setting pred name {pred.name} score to {response.data}")


if __name__ == "__main__":
    # Example of using EvalFnToTextLoss
    from adalflow.utils import setup_env, get_logger
    from adalflow.eval.answer_match_acc import AnswerMatchAcc
    from adalflow.components.model_client import OpenAIClient  # dir: model_client
    from adalflow.core.generator import Generator, BackwardEngine
    from adalflow.core.component import func_to_data_component

    logger = get_logger(level="DEBUG", filename="lib_text_grad.log")

    setup_env()

    gpt_3_model = {
        "model_client": OpenAIClient(),
        "model_kwargs": {
            "model": "gpt-3.5-turbo",
        },
    }
    gpt_4o_model = {
        "model_client": OpenAIClient(),
        "model_kwargs": {
            "model": "gpt-4o",
        },
    }

    @func_to_data_component
    def parse_integer_answer(answer: str, only_first_line: bool = False):
        try:
            if only_first_line:
                answer = answer.strip().split("\n")[0]
            answer = answer.strip()
            # find the last token that has a number in it
            answer = [
                token for token in answer.split() if any(c.isdigit() for c in token)
            ][-1]
            answer = answer.split(".")[0]
            answer = "".join([c for c in answer if c.isdigit()])
            answer = int(answer)

        except (ValueError, IndexError):
            answer = 0

        return answer

    evaluator = AnswerMatchAcc()
    eval_fn_desc = "Answer Match Accuracy"
    single_compute_eval = AnswerMatchAcc().compute_single_item

    backward_engine = BackwardEngine(**gpt_4o_model)
    # backward_engine.set_mock_output(mock_output_data="1")

    eval_fn_to_text_loss = EvalFnToTextLoss(
        eval_fn=single_compute_eval,
        eval_fn_desc=eval_fn_desc,
        backward_engine=backward_engine,
    )
    x = Parameter(
        name="x",
        data="I have a cauliflower, a stalk of celery, a cabbage, and a garlic. How many vegetables do I have?",
        requires_opt=False,
        role_desc="The question to the language model",
    )

    system_prompt = Parameter(
        name="system_prompt",
        data="You will answer a reasoning question. Think step by step. The last line of your response should be of the following format: 'Answer: $VALUE' where VALUE is a numerical value.",
        requires_opt=True,
        role_desc="structured system prompt to a somewhat capable language model that specifies the behavior and strategies for the QA task",
    )

    model = Generator(
        prompt_kwargs={"task_desc_str": system_prompt},
        **gpt_3_model,
        output_processors=parse_integer_answer,
    )
    # model.set_mock_output(mock_output_data="4")
    model.train()

    y: Parameter = model(prompt_kwargs={"input_str": x})

    loss = eval_fn_to_text_loss(
        {
            "y": y,
            "y_gt": Parameter(
                name="y_gt",
                data="4",
                requires_opt=False,
                role_desc="Correct answer",
            ),
        }
    )
    loss.backward()
    assert len(loss.predecessors) == 2
    assert len(y.predecessors) == 2
    dot = loss.draw_graph(add_grads=True, filepath="real_data")
    # print("dot: ", dot)

#     Variable(data=1, requires_opt=True, role_desc=Output of the string-based function with purpose:
#              The runtime of string-based function that checks if the prediction is correct.,
#              predecessors={Variable(data=4, requires_opt=False, role_desc=correct answer for the query,
#                                     predecessors=set(), gradients=set()),
#                                     Variable(data=To determine the number of vegetables,
#                                              we need to count each individual item.
#                                                The cauliflower, celery, cabbage,
#                                                and garlic are all separate vegetables.
#                                                Therefore, you have 4 vegetables in total.

# Answer: 4, requires_opt=True, role_desc=response from the language model,
# predecessors={Variable(data=I have a cauliflower, a stalk of celery, a cabbage, and a garlic. How many vegetables do I have?, requires_opt=False, role_desc=query to the language model, predecessors=set(), gradients=set()), Variable(data=You will answer a reasoning question. Think step by step. The last line of your response should be of the following format: 'Answer: $VALUE' where VALUE is a numerical value., requires_opt=True, role_desc=structured system prompt to a somewhat capable language model that specifies the behavior and strategies for the QA task, predecessors=set(), gradients=set())}, gradients=set())}, gradients=set())

# loss: Parameter(alias=None, data=1.0, requires_opt=True, role_desc=Output of EvalFnToTextLoss with eval_fn_desc: Answer Match Accuracy,
# predecessors={Parameter(alias=None, data=1, requires_opt=False, role_desc=Predicted answer, predecessors=set(), gradients=set()),
# Parameter(alias=None, data=1, requires_opt=False, role_desc=Correct answer, predecessors=set(), gradients=set())}, gradients=set())
# {'alias': None, 'data': 1.0, 'requires_opt': True, 'role_desc': 'Output of EvalFnToTextLoss with eval_fn_desc: Answer Match Accuracy', 'predecessors': [{'alias': None, 'data': '1', 'requires_opt': False, 'role_desc': 'Predicted answer', 'predecessors': [], 'gradients': [], 'proposed_data': None, 'gradients_context': [], 'grad_fn': 'None'}, {'alias': None, 'data': '1', 'requires_opt': False, 'role_desc': 'Correct answer', 'predecessors': [], 'gradients': [], 'proposed_data': None, 'gradients_context': [], 'grad_fn': 'None'}], 'gradients': [], 'proposed_data': None, 'gradients_context': [], 'grad_fn': '__main__.EvalFnToTextLoss.backward'}



from typing import List

from adalflow.core.component import Sequential
from adalflow.core.types import Document

from adalflow.utils.config import new_component
from adalflow.database.sqlalchemy.sqlachemy_manager import DatabaseManager
from adalflow.database.sqlalchemy.model import DocumentModel
from adalflow.database.sqlalchemy.pipeline import default_config


# TODO:  async call
class EmbeddingPipeline:
    __doc__ = r"""Pipeline to process documents and store them in the database

    Args:
        batch_size (int, optional): Batch size for processing. Defaults to 100.

    Example:

    ..code-block:: python

        from adalflow.core.types import Document
        from adalflow.utils import setup_env  # noqa

        documents = [
            {
                "meta_data": {"title": "Li Yin's profile"},
                "text": "My name is Li Yin, I love rock climbing"
                + "lots of nonsense text" * 500,
                "id": "doc1",
            },
            {
                "meta_data": {"title": "Interviewing Li Yin"},
                "text": "lots of more nonsense text" * 250
                + "Li Yin is a software developer and AI researcher"
                + "lots of more nonsense text" * 250,
                "id": "doc2",
            },
        ]
        db_name = "vector_db"
        postgres_url = f"postgresql://postgres:password@localhost:5432/{db_name}"
        ep = EmbeddingPipeline()
        ep.setup_database_manager(postgres_url)

        new_documents = [Document(**doc) for doc in documents]

        ep(new_documents)

    Note:
     - Before you run the pipeline, ensure you have created the database using the `create_tables.py` script
    """

    def __init__(self, batch_size: int = 100):
        self.batch_size = batch_size

        self.config_dict = default_config
        self.document_splitter = new_component(self.config_dict["document_splitter"])
        self.to_embeddings = new_component(self.config_dict["to_embeddings"])
        self.vectorizer = self.to_embeddings.vectorizer

        self.data_transformer = Sequential(
            self.document_splitter,
            self.to_embeddings,
        )

    def setup_database_manager(self, db_path: str):
        self.db_manager = DatabaseManager(db_path)

    def process_batch(self, documents: List[Document]):
        transformed_documents = self.data_transformer(documents)
        session = self.db_manager.get_session()
        documents_to_upload = []
        for doc in transformed_documents:
            item = doc.to_dict()
            if "score" in item:
                del item["score"]
            documents_to_upload.append(item)

        DocumentModel.insert_update_bulk(documents_to_upload, session)
        return transformed_documents

    def __call__(self, documents: List[Document]):
        batch_size = self.batch_size
        for i in range(0, len(documents), batch_size):

            List = documents[i : i + batch_size]
            print(i, len(List))
            self.process_batch(List)


if __name__ == "__main__":
    from adalflow.core.types import Document
    from adalflow.utils import setup_env  # noqa

    documents = [
        {
            "meta_data": {"title": "Li Yin's profile"},
            "text": "My name is Li Yin, I love rock climbing"
            + "lots of nonsense text" * 500,
            "id": "doc1",
        },
        {
            "meta_data": {"title": "Interviewing Li Yin"},
            "text": "lots of more nonsense text" * 250
            + "Li Yin is a software developer and AI researcher"
            + "lots of more nonsense text" * 250,
            "id": "doc2",
        },
    ]
    db_name = "vector_db"
    postgres_url = f"postgresql://postgres:password@localhost:5432/{db_name}"
    ep = EmbeddingPipeline()

    ep.setup_database_manager(postgres_url)
    # new_documents = [Document(**doc) for doc in documents]
    # ep(new_documents)

    db_manager = DatabaseManager(postgres_url)
    tables = db_manager.list_table_schemas()
    print(tables)
    column = "vector"
    table_name = "document"
    vectorizer = ep.vectorizer
    query = "What did the author do?"
    query_embedding = vectorizer(query).data[0].embedding

    def format_query_str(
        table_name: str, column: str, query_embedding: List[float], top_k: int = 5
    ):
        return f"""SELECT * FROM {table_name} ORDER BY {column} <-> '{str(query_embedding)}' LIMIT {top_k};"""

    l2_nearest_neighbor = format_query_str(table_name, column, query_embedding)
    print(l2_nearest_neighbor)
    results = db_manager.execute_query(l2_nearest_neighbor)
    print(results)



import hashlib
import diskcache as dc
from pathlib import Path
from typing import Union


def hash_text(text: str):
    return hashlib.sha256(f"{text}".encode()).hexdigest()


def hash_text_sha1(text: str):  # 160 bits
    return hashlib.sha1(text.encode()).hexdigest()


def direct(text: str):
    return text


class CachedEngine:
    def __init__(self, cache_path: Union[str, Path]):
        super().__init__()
        self.cache_path = Path(cache_path)
        self.cache_path.parent.mkdir(parents=True, exist_ok=True)

        self.cache = dc.Cache(cache_path)

    def _check_cache(self, prompt: str):
        hash_key = hash_text(prompt)
        if hash_key in self.cache:
            return self.cache[hash_key]
        else:
            return None

    def _save_cache(self, prompt: str, response: str):
        hash_key = hash_text(prompt)
        self.cache[hash_key] = response

    def __getstate__(self):
        # Remove the cache from the state before pickling
        state = self.__dict__.copy()
        del state["cache"]
        return state

    def __setstate__(self, state):
        # Restore the cache after unpickling
        self.__dict__.update(state)
        self.cache = dc.Cache(self.cache_path)



from .trainer import Trainer
from .adal import AdalComponent
from adalflow.utils.registry import EntityMapping


__all__ = [
    "Trainer",
    "AdalComponent",
]
for name in __all__:
    EntityMapping.register(name, globals()[name])



default_config = {
    "document_splitter": {
        "component_name": "DocumentSplitter",
        "component_config": {
            "split_by": "word",
            "split_length": 500,
            "split_overlap": 200,
        },
    },
    "to_embeddings": {
        "component_name": "ToEmbeddings",
        "component_config": {
            "embedder": {
                "component_name": "Embedder",
                "component_config": {
                    "model_client": {
                        "component_name": "OpenAIClient",
                        "component_config": {},
                    },
                    "model_kwargs": {
                        "model": "text-embedding-3-small",
                        "dimensions": 256,
                        "encoding_format": "float",
                    },
                },
            },
            "batch_size": 100,
        },
    },
}



"""Text-grad operations such as Sum and Aggregate."""

from typing import List
import logging

from adalflow.optim.function import BackwardContext
from adalflow.optim.parameter import Parameter, OutputParameter
from adalflow.optim.types import ParameterType
from adalflow.optim.grad_component import GradComponent

log = logging.getLogger(__name__)


def sum_ops(params: List[Parameter]) -> Parameter:
    """
    Represents a sum operation on a list of variables.
    In TextGrad, sum is simply concatenation of the values of the variables.

    :param variables: The list of variables to be summed (concatenated).
    :type variables: List[Variable]
    :return: A new variable representing the sum of the input variables.
    :rtype: Variable
    """
    for param in params:
        if not isinstance(param, Parameter):
            raise ValueError(
                f"Sum operation only accepts a list of Parameters, got {type(param)}"
            )
    return Sum()(params)


class Sum(GradComponent):
    __doc__ = """The class to define a sum operation on a list of parameters, such as losses or gradients.

    It enables gradients combination of a batch of data samples.
    """

    name = "Sum"

    def __init__(self):
        super().__init__(desc="Sum a list of parameters")

    def __call__(self, *args, **kwargs):
        return self.forward(*args, **kwargs)

    def forward(self, params: List[Parameter]) -> Parameter:
        """
        Performs the forward pass of the sum operation.
        This is a simple operation that concatenates the values of the parameters.

        :param params: The list of parameters to be summed.
        :type params: List[Parameter]
        :rtype: Parameter
        """
        for param in params:
            if not isinstance(param, Parameter):
                raise ValueError(
                    f"Sum operation only accepts a list of Parameters, got {type(param)}"
                )
        concat_values = ",".join([str(p.data) for p in params])  # default concatenation
        role_descriptions = set([p.role_desc for p in params])
        role_descriptions = ", ".join(role_descriptions)

        total = OutputParameter(
            data=concat_values,
            role_desc=f"A combination of a list of variables: {role_descriptions}",
            requires_opt=any([p.requires_opt for p in params]),
            name="sum",
            score=sum([p.score for p in params]),  # total has a score
            param_type=ParameterType.SUM_OUTPUT,
        )
        total.set_predecessors(params)
        total.trace_forward_pass(
            input_args=params,
            full_response=concat_values,
            id=total.id,
            name=total.name,
        )

        log.info("Sum forward", extra={"total": total.data})

        total.set_grad_fn(
            BackwardContext(
                backward_fn=self.backward,
                summation=total,
                disable_backward_engine=False,
            )
        )

        return total

    def backward(self, summation: Parameter, *args, **kwargs):
        """
        Computes gradients for the predecessors of the sum operation.
        There is no gradient computation for the sum operation itself.
        It is a simple way to combine multiple losses for convenience.

        sum.backward() => [loss1.backward(), loss2.backward(), ...]

        :param summation: The parameter representing the sum.
        :type summation: Parameter
        """
        log.info(f"Sum backward: {summation.data}")
        pred_params = summation.predecessors  # losses
        summation_gradients = summation.get_gradient_and_context_text().strip()
        for param in pred_params:

            if param.check_if_already_computed_gradient_respect_to(summation.id):
                log.info(
                    f"Gradient already computed for {param.role_desc} with respect to {summation.role_desc}"
                )
                print(
                    f"Gradient already computed for {param.role_desc} with respect to {summation.role_desc}"
                )
                continue

            # add a combined gradients
            if (
                summation_gradients == ""
            ):  # as loss sum to be the base, it simply allows gradients computations on multiple losses
                param_gradient_value = ""
            else:  # as a mid layer, it will have a combined feedback

                param_gradient_value = f"Here is the combined feedback we got for this specific {param.role_desc} and other parameters: {summation_gradients}."

            extra = {
                "p_gradient_value": param_gradient_value,
                "summation_role": summation.role_desc,
            }
            log.info(f"""Idempotent sum backward: {extra}""")


if __name__ == "__main__":
    # test the sum ops

    a = Parameter(data=1)
    b = Parameter(data=2)
    c = sum_ops(List[a, b])
    c.backward()



"""GradientContext and Gradient"""

import uuid
from typing import Any, Optional, TYPE_CHECKING

if TYPE_CHECKING:
    from adalflow.optim.parameter import Parameter
from dataclasses import dataclass, field
from adalflow.core.base_data_class import DataClass

__all__ = ["GradientContext", "Gradient"]


@dataclass
class GradientContext(DataClass):
    """GradientContext is used to describe the component's function and trace its input and output.

    To get the component's function desc, use GradientContext.to_yaml_signature()
    To get the data: use instance.to_yaml()
    """

    variable_desc: str = field(
        metadata={"desc": "The description of the target parameter"}
    )

    input_output: str = field(
        metadata={
            "desc": "The context of the gradient in form of a conversation indicating \
                the relation of the current parameter to the response parameter"
        }
    )
    response_desc: str = field(
        metadata={"desc": "The description of the response parameter"}
    )
    # input: Dict[str, Any] = field(
    #     metadata={"desc": "The input to the whole system"}, default=None
    # )

    # ground_truth: Any = field(
    #     metadata={"desc": "The ground truth of the response parameter"}, default=None
    # )


@dataclass
class Gradient(DataClass):
    __doc__ = r"""It will handle gradients and feedbacks.

    It tracks the d_from_response_id / d_to_pred_id and the score of the whole response.

    if two gradients have the same data_id, different from_response_id, and same from_response_component_id, this is a cycle component structure.
    """
    data_id: Optional[str] = None  # the id of the response from data in the dataset
    from_response_component_id: str = (
        None  # the id of the component from which the gradient is calculated
    )
    order: Optional[int] = None  # the order of the gradient in the list of gradients

    from_response_id: str = (
        None  # the id of the response from which the gradient is calculated
    )

    to_pred_id: str = (
        None  # the id of the parameter to which the gradient is calculated and attached to d(from_response_id) / d(to_pred_id)
    )

    score: Optional[float] = None

    context: GradientContext = None
    data: Any = None
    prompt: Optional[str] = None  # the LLM prompt to generate the gradient

    is_default_copy: bool = False  # whether the gradient is a default copy

    def __init__(
        self,
        *,
        from_response: "Parameter",
        to_pred: "Parameter",
        id: Optional[str] = None,  # the id of the gradient
        score: Optional[float] = None,
        data_id: Optional[str] = None,
        data: Any = None,
    ):
        self.id = id or str(uuid.uuid4())
        self._generate_name(from_response, to_pred)
        self.from_response_component_id = from_response.component_trace.id
        if not self.from_response_component_id:
            raise ValueError(
                "The from_response_component_id should not be None. Please ensure the component_trace is set."
            )
        self.from_response_id = from_response.id
        self.to_pred_id = to_pred.id
        self.score = score
        self.data_id = data_id
        if self.data_id is None:
            raise ValueError("The data_id should not be None.")
        self.data = data
        self.order = None

    def _generate_name(self, response: "Parameter", pred: "Parameter"):
        self.name = f"d_{response.name}_/_{pred.name}({response.id}_/_{pred.id})"
        self.role_desc = f"Gradient from {response.name} to {pred.name}"

    def add_context(self, context: GradientContext):
        self.context = context

    def add_data(self, data: Any):
        self.data = data

    def update_from_to(self, from_response: "Parameter", to_pred: "Parameter"):
        self.from_response_id = from_response.id
        self.to_pred_id = to_pred.id
        self._generate_name(from_response, to_pred)
        self.from_response_component_id = from_response.component_trace.id

    def add_prompt(self, prompt: str):
        self.prompt = prompt

    def __hash__(self):
        # Use immutable and unique attributes to compute the hash
        return hash((self.id, self.data_id, self.from_response_id, self.to_pred_id))

    def __eq__(self, other):
        # Ensure equality comparison is based on the same unique attributes
        if not isinstance(other, Gradient):
            return False
        return (
            self.id == other.id
            and self.data_id == other.data_id
            and self.from_response_id == other.from_response_id
            and self.to_pred_id == other.to_pred_id
        )



"""Base class for Autograd Components that can be called and backpropagated through."""

from typing import TYPE_CHECKING, Callable, Optional, Dict
from collections import OrderedDict
import uuid
import logging
from copy import deepcopy

if TYPE_CHECKING:
    from adalflow.core.generator import BackwardEngine

    from adalflow.core import ModelClient
from adalflow.optim.parameter import (
    Parameter,
    OutputParameter,
)
from adalflow.optim.gradient import Gradient, GradientContext
from adalflow.optim.types import ParameterType
from adalflow.core.types import GeneratorOutput
from adalflow.utils import printc

import json

from adalflow.core.component import Component
from adalflow.optim.function import BackwardContext
from adalflow.core.prompt_builder import Prompt
from adalflow.optim.text_grad.backend_engine_prompt import (
    GRAD_COMPONENT_CONVERSATION_TEMPLATE_STRING,
    LOSS_CONVERSATION_START_INSTRUCTION_STRING_FN,
    OBJECTIVE_INSTRUCTION_BASE,
    OBJECTIVE_INSTRUCTION_CHAIN,
)


__all__ = ["GradComponent", "FunGradComponent", "fun_to_grad_component"]

log = logging.getLogger(__name__)


class GradComponent(Component):
    __doc__ = """A base class to define interfaces for an auto-grad component/operator.

    Compared with `Component`, `GradComponent` defines three important interfaces:
    - `forward`: the forward pass of the function, returns a `Parameter` object that can be traced and backpropagated.
    - `backward`: the backward pass of the function, updates the gradients/prediction score backpropagated from a "loss" parameter.
    - `set_backward_engine`: set the backward engine(a form of generator) to the component, which is used to backpropagate the gradients using LLM.

    The __call__ method will check if the component is in training mode,
    and call the `forward` method to return a `Parameter` object if it is in training mode,
    otherwise, it will call the `call` method to return the output such as "GeneratorOutput", "RetrieverOutput", etc.

    Note: Avoid using the attributes and methods that are defined here and in the `Component` class unless you are overriding them.
    """

    backward_engine: "BackwardEngine"
    _component_type = "grad"
    id = None
    _component_desc = "GradComponent"
    _disable_backward_engine = False

    def __init__(
        self,
        desc: str,  # what is this component for? Useful for training.
        name: Optional[str] = None,
        backward_engine: Optional["BackwardEngine"] = None,
        model_client: "ModelClient" = None,
        model_kwargs: Dict[str, object] = None,
    ):
        super().__init__()
        super().__setattr__("id", str(uuid.uuid4()))

        self.desc = desc
        self.backward_engine = backward_engine
        self.model_client = model_client
        self.name = name or f"{self.__class__.__name__}"

        self.backward_engine = None
        if backward_engine is None:
            log.info(
                "EvalFnToTextLoss: No backward engine provided. Creating one using model_client and model_kwargs."
            )
            if model_client and model_kwargs:

                self.set_backward_engine(backward_engine, model_client, model_kwargs)
        else:
            if not isinstance(backward_engine, BackwardEngine):
                raise TypeError(
                    "EvalFnToTextLoss: backward_engine must be an instance of BackwardEngine."
                )
            self.backward_engine = backward_engine

    def set_backward_engine(
        self,
        backward_engine: "BackwardEngine" = None,
        model_client: "ModelClient" = None,
        model_kwargs: Dict[str, object] = None,
    ):
        from adalflow.core.generator import BackwardEngine

        self.backward_engine = backward_engine
        if not backward_engine:
            log.info(
                "EvalFnToTextLoss: No backward engine provided. Creating one using model_client and model_kwargs."
            )
            self.backward_engine = BackwardEngine(model_client, model_kwargs)
        else:
            if type(backward_engine) is not BackwardEngine:
                raise TypeError(
                    f"EvalFnToTextLoss: backward_engine must be an instance of BackwardEngine. Got {type(backward_engine)}."
                )

    def disable_backward_engine(self):
        r"""Does not run gradients generation, but still with backward to gain module-context"""
        self._disable_backward_engine = True

    def call(self, *args, **kwargs):
        raise NotImplementedError("call method is not implemented")

    async def acall(self, *args, **kwargs):
        r"""Implement this for your async call."""
        raise NotImplementedError("acall method is not implemented")

    def forward(self, *args, **kwargs) -> "Parameter":
        r"""Default forward method for training:
        1. for all args and kwargs, if it is a `Parameter` object, it will be tracked as `Predecessor`.
        2. Trace input_args and full_response in the parameter object.
        3. Return the parameter object.
        """

        from adalflow.optim.parameter import Parameter, OutputParameter

        log.debug(
            f"Forwarding through {self.name} with args: {args} and kwargs: {kwargs}"
        )

        # 1. get all predecessors from all args and kwargs
        input_args = OrderedDict()

        # Add positional args to the ordered dict
        for idx, arg in enumerate(args):
            input_args[f"arg_{idx}"] = arg

        # Get data id from the kwargs
        data_id = kwargs.get("id", None)

        # Add keyword args to the ordered dict, preserving order
        predecessors = []
        for v in input_args.values():
            if isinstance(v, Parameter):
                predecessors.append(v)
                if v.param_type == ParameterType.INPUT:
                    v.data_id = kwargs.get("id", None)
                if data_id is None:
                    data_id = v.data_id
        # printc(f"kwargs: {kwargs}")
        # discard_keys = []
        for k, v in kwargs.items():
            if isinstance(v, Parameter):
                predecessors.append(v)
                if v.param_type == ParameterType.INPUT:
                    v.data_id = kwargs.get("id", None)
                if data_id is None:
                    data_id = v.data_id
            # support list of Parameters by flattening them
            elif isinstance(v, list):
                for i, p in enumerate(v):
                    if isinstance(p, Parameter):
                        predecessors.append(p)

        # 2. unwrap the parameter object to take only the data, successor_map_fn: lambda x: x.data in default
        # unwrap args
        unwrapped_args = []
        for k, v in input_args.items():
            if isinstance(v, Parameter):
                unwrapped_args.append(v.map_to_successor(self))
            else:
                unwrapped_args.append(v)

        unwrapped_kwargs = {}
        # unwrap kwargs
        for k, v in kwargs.items():
            if isinstance(v, Parameter):
                unwrapped_kwargs[k] = v.map_to_successor(self)
            elif isinstance(v, list):
                values = []
                for p in v:
                    if isinstance(p, Parameter):
                        values.append(p.map_to_successor(self))
                    else:
                        values.append(p)
                unwrapped_kwargs[k] = values
            else:
                unwrapped_kwargs[k] = v

        # 3. call the function with unwrapped args and kwargs
        unwrapped_args = tuple(unwrapped_args)

        log.debug(f"Unwrapped args: {unwrapped_args}")
        log.debug(f"Unwrapped kwargs: {unwrapped_kwargs}")

        call_response = self.call(*unwrapped_args, **unwrapped_kwargs)

        if isinstance(call_response, Parameter):
            raise ValueError(
                f"A GradComponent call should not return Parameter, got {call_response.name}"
            )
            predecessors.append(call_response)
            return call_response

        # 4. Create a Parameter object to trace the forward pass
        # use unwrapped args  and unwrapped kwargs to trace the forward pass
        tracing_args = {i: v for i, v in enumerate(unwrapped_args)}
        tracing_args.update(**unwrapped_kwargs)

        response = OutputParameter(
            data=call_response,
            name=self.name + "_output",
            role_desc=self.name + " response",
            param_type=ParameterType.OUTPUT,
            data_id=data_id,
        )
        response.set_predecessors(predecessors)
        response.trace_forward_pass(
            input_args=tracing_args,
            full_response=call_response,
            id=self.id,  # this is component id
            name=self.name,
        )
        response.set_grad_fn(
            BackwardContext(
                backward_fn=self.backward,
                response=response,
                id=data_id,
                input_kwargs=kwargs,
                disable_backward_engine=self._disable_backward_engine,
            )
        )
        return response

    def backward_with_pass_through_gradients(
        self, *, response: "Parameter", id: str = None, **kwargs
    ):
        """Pass-through gradient to the predecessors.

        Backward pass of the function. In default, it will pass all the scores to the predecessors.

        Note: backward is mainly used internally and better to only allow kwargs as the input.

        Subclass should implement this method if you need additional backward logic.
        """

        log.info(f"GradComponent backward: {response.name}")
        children_params = response.predecessors

        if response.get_gradient_and_context_text().strip() == "":
            log.info(f"Generator: Backward: No gradient found for {response}.")

        for _, pred in enumerate(children_params):
            if response.score is not None:
                pred.set_score(response.score)

            if pred.param_type == ParameterType.DEMOS:
                pred.add_score_to_trace(
                    trace_id=id, score=response.score, is_teacher=self.teacher_mode
                )
            for grad in response.gradients:
                # NOTE: make a copy of the gradient, we should not modify the original gradient
                grad = deepcopy(grad)

                grad.is_default_copy = (
                    True  # response and pred will keep the original gradient
                )
                pred.add_gradient(grad)

    @staticmethod
    def _backward_through_one_predecessor(
        pred: Parameter,
        kwargs: Dict[str, Parameter],
        response: Parameter,
        desc: str,
        backward_engine: "BackwardEngine",
        ground_truth: object = None,
        is_intermediate_node: bool = False,  # if the node is an intermediate node in the backpropagation chain
        metadata: Dict[str, str] = None,
        disable_backward_engine: bool = False,
    ):
        if not pred.requires_opt:
            if response.score is not None:
                pred.set_score(response.score)
            log.debug(
                f"EvalFnToTextLoss: Skipping {pred} as it does not require optimization."
            )
            return
        log.debug(
            f"EvalFnToTextLoss: Backward through {pred}, is_intermediate_node: {is_intermediate_node}"
        )

        if pred.check_if_already_computed_gradient_respect_to(response.id):
            log.info(
                f"EvalFnToTextLoss: Gradient already computed for {pred.role_desc} with respect to {response.role_desc}"
            )

            return

        if backward_engine is None:
            log.error(
                "EvalFnToTextLoss: backward_engine is required for text prompt optimization."
            )
            raise ValueError(
                "EvalFnToTextLoss: backward_engine is required for text prompt optimization."
            )

        instruction_str, objective_str = None, None

        # convert kwargs to key, (value, type(eval_input))

        inputs = {}

        for k, v in kwargs.items():
            if isinstance(v, Parameter):
                inputs[k] = (v.get_param_info(), str(type(v.eval_input)))
            elif isinstance(v, list):
                # flat the list to multiple parameters

                for i, p in enumerate(v):
                    if isinstance(p, Parameter):
                        flat_key = f"{k}_{i}"
                        inputs[flat_key] = (p.get_param_info(), str(type(p.eval_input)))

        # response information
        conversation_str = Prompt(
            GRAD_COMPONENT_CONVERSATION_TEMPLATE_STRING,
            prompt_kwargs={
                "inputs": inputs,
                "component_desc": desc,
                "response_value": response.get_prompt_data(),
                "metadata": json.dumps(metadata) if metadata else None,
            },
        )()

        conv_ins_template = LOSS_CONVERSATION_START_INSTRUCTION_STRING_FN
        obj_ins_template = OBJECTIVE_INSTRUCTION_BASE

        if is_intermediate_node:
            printc(f"is_intermediate_node: {is_intermediate_node}")
            # conv_ins_template = CONVERSATION_START_INSTRUCTION_STRING_FN_CHAIN
            obj_ins_template = OBJECTIVE_INSTRUCTION_CHAIN

        instruction_str = Prompt(
            conv_ins_template,
            prompt_kwargs={
                "variable": pred.get_param_info(),
                "conversation_str": conversation_str,
            },
        )()
        response_gradient = response.get_gradients_str()
        # response_gradient = response.get_gradients_component_schema()
        if not response_gradient:
            raise ValueError(
                f"Generator: No gradient found for {response}. Please check the response. pred: {pred}"
            )
        objective_str = Prompt(
            obj_ins_template,
            prompt_kwargs={
                "response_name": response.name,
                "response_desc": response.role_desc,
                "response_gradient": response_gradient,
            },
        )()

        log.info(f"EvalFnToTextLoss: Instruction: {instruction_str}")
        log.info(f"EvalFnToTextLoss: Objective: {objective_str}")
        log.info(f"EvalFnToTextLoss: Conversation: {conversation_str}")

        # Compute the gradient
        backward_engine_prompt_kwargs = {
            "conversation_sec": instruction_str,
            "objective_instruction_sec": objective_str,
        }
        gradient_value_data = None
        if not disable_backward_engine:
            gradient_value: GeneratorOutput = backward_engine(
                prompt_kwargs=backward_engine_prompt_kwargs
            )
            gradient_prompt = backward_engine.get_prompt(
                **backward_engine_prompt_kwargs
            )
            gradient_value_data = (
                gradient_value.data
                or backward_engine.failure_message_to_optimizer(
                    gradient_response=gradient_value
                )
            )

            gradient_value_data = (
                f"expected answer: {ground_truth},\n Feedback: {gradient_value_data}"
            )

            log.debug(f"EvalFnToTextLoss: Gradient for {pred}: {gradient_value_data}")

        gradient_param = Gradient(
            data=gradient_value_data,
            data_id=response.data_id,
            score=response.score,
            from_response=response,
            to_pred=pred,
        )
        gradient_param.add_prompt(gradient_prompt)
        gradient_param.add_context(
            GradientContext(
                input_output=conversation_str,
                response_desc=response.role_desc,
                variable_desc=pred.role_desc,
            )
        )
        pred.add_gradient(gradient_param)

        if response.score is not None:
            pred.set_score(response.score)
        pred.set_gt(ground_truth)

    def backward(
        self,
        *,
        response: "OutputParameter",
        id: str = None,
        disable_backward_engine=False,
        **kwargs,
    ):
        """Backward pass of the function. In default, it will pass all the scores to the predecessors.

        Note: backward is mainly used internally and better to only allow kwargs as the input.

        Subclass should implement this method if you need additional backward logic.
        """

        log.info(f"GradComponent backward: {response.name}")
        children_params = response.predecessors

        input_kwargs = kwargs.get("input_kwargs", {})

        is_intermediate_node = False
        response_gradient_context = response.get_gradient_and_context_text().strip()
        if response_gradient_context != "":
            log.info("EvalFnToTextLoss is an intermediate node.")
            is_intermediate_node = True

        if response.get_gradient_and_context_text().strip() == "":
            log.info(f"Generator: Backward: No gradient found for {response}.")

        # use pass through gradient when there is one predecessor
        if not self.backward_engine or len(children_params) < 2:
            self.backward_with_pass_through_gradients(response=response, id=id)

        else:

            for _, pred in enumerate(children_params):
                if response.score is not None:
                    pred.set_score(response.score)
                printc(f"score {response.score} for pred name: {pred.name}")
                if not pred.requires_opt:
                    continue

                if pred.param_type == ParameterType.DEMOS:
                    pred.add_score_to_trace(
                        trace_id=id, score=response.score, is_teacher=self.teacher_mode
                    )

                printc(f"pred: {pred.name}, response: {response.name}")

                self._backward_through_one_predecessor(
                    pred=pred,
                    kwargs=input_kwargs,
                    response=response,
                    backward_engine=self.backward_engine,
                    desc=self.desc,
                    is_intermediate_node=is_intermediate_node,
                    disable_backward_engine=disable_backward_engine,
                )


class FunGradComponent(GradComponent):

    def __init__(
        self,
        fun: Optional[Callable] = None,
        afun: Optional[Callable] = None,
        desc: str = "",
        doc_string=None,
    ):
        desc = desc or fun.__doc__ or f"Function: {fun.__name__}"

        super().__init__(desc=desc, name=fun.__name__)
        self.fun_name = fun.__name__
        self.fun = fun
        # set the docstring
        self.doc_string = doc_string
        setattr(
            self.fun,
            "__doc__",
            doc_string or fun.__doc__ or f"Function: {fun.__name__}",
        )

        setattr(self.fun, "__name__", fun.__name__)

    def call(self, *args, **kwargs):

        kwargs.pop("doc_string", None)

        return self.fun(*args, **kwargs)

    def forward(self, *args, **kwargs) -> Parameter:
        """add func_doc_string to the kwargs before calling the super().forward"""
        kwargs["doc_string"] = self.doc_string
        output = super().forward(*args, **kwargs)
        return output

    def _extra_repr(self) -> str:
        return (
            super()._extra_repr()
            + f"fun_name={self.fun_name}, fun={self.fun.__name__}, fun_doc={self.fun.__doc__}"
        )


def fun_to_grad_component(desc: str = "", doc_string=None) -> Callable:
    """
    Return a decorator that, when applied to a function `fun`,
    wraps it in a GradComponent with the given `desc`.

    Examples:

    1. As a decorator:


    ::code-block :: python

        @fun_to_grad_component(desc="This is a test function", doc_string=Parameter(
            data="Finish the task with verbatim short factoid responses from retrieved context.",
            param_type=ParameterType.PROMPT,
            requires_opt=True,
            role_desc="Instruct how the agent creates the final answer from the step history.",
        ))
        def my_function(x):
            return x + 1

        print(my_function(1))

    2. As a function:

    ::code-block :: python

        def my_function(x):
            return x + 1

        my_function_component = fun_to_grad_component(desc="This is a test function")(my_function)
    """

    def decorator(fun):
        # 1) build the class name
        class_name = (
            "".join(part.capitalize() for part in fun.__name__.split("_"))
            + "GradComponent"
        )
        # 2) define the new class
        printc(f"fun doc 1: {doc_string}, desc: {desc}")
        component_class = type(
            class_name,
            (FunGradComponent,),
            {
                "__init__": lambda self: FunGradComponent.__init__(
                    self, fun=fun, desc=desc, doc_string=doc_string
                )
            },
        )
        return component_class()

    return decorator


if __name__ == "__main__":
    # Test FunGradComponent
    from adalflow.optim.parameter import Parameter

    def my_function(x):
        __doc__ = Parameter(  # noqa F841
            data="Finish the task with verbatim short factoid responses from retrieved context.",
            param_type=ParameterType.PROMPT,
            requires_opt=True,
            role_desc="Instruct how the agent creates the final answer from the step history.",
        )
        return x + 1

    my_function_component = fun_to_grad_component()(my_function)
    print(my_function_component)  # 2
    # eval mode
    output = my_function_component(1)
    print(output)
    # training mode
    my_function_component.train()
    output = my_function_component(Parameter(data=1, name="input"))
    print(output)

    # now test the decorator
    @fun_to_grad_component(
        desc="This is a test function",
        doc_string=Parameter(
            data="Finish the task with verbatim short factoid responses from retrieved context.",
            param_type=ParameterType.PROMPT,
            requires_opt=True,
            role_desc="Instruct how the agent creates the final answer from the step history.",
        ),
    )
    def my_function(x):

        return x + 1

    print(my_function(1))
    # eval mode
    output = my_function(1)
    print(output)
    assert output == 2

    # training mode
    my_function.train()
    print(my_function)
    output = my_function(Parameter(data=1, name="input"))
    print(output)



"""Meta-prompts for the backward engine.

Optimized from Textual Auto-diff and enhanced with peer variables.

Reference: TextGrad: Automatic “Differentiation” via Text."""

# NOTE: not receive feedback for good performing case is important for performance
# NOTE: having peers is important to keep the scope of the prompt consistent and not cross-reference with other variables
### System prompt and the template is shared by all GradComponent ###


FEEDBACK_ENGINE_TEMPLATE = r"""<START_OF_SYSTEM_PROMPT>
You MUST determining the root cause of a system error.
You start with an evaluation function that measures performance, and you receive the system input.
The system can be a a compound system, potentially consisting of multiple components.
You work on one component.
You will receive feedback from your direct successor component, and your goal is to investigate your component’s inputs and outputs to identify whether any of your input variables are causing the error.

Your target input variable is enclosed in <TARGET_VARIABLE> (representing one of the input variables that may or may not be causing the error).
Alternatively, it may be enclosed in <VARIABLES> tags (in which case you must pass feedback to all variables, indicating which ones cause the errors and which do not).

1. From <CONVERSATION></CONVERSATION> section, you can find how the variable is obtained and used.
2. As there might be multiple precedessors, and multi-components, it is possible that the feedback/error is not directly related to the variable itself.
3. When you reason, really think about the variable's role in the component(infer from the CONVERSATION section) and the VARIABLE section before you provide feedback.
4. Be specific, concise, critical, and direct.
5. Maximum 3 sentences.

[Cycle]: If the same DataID has multiple gradients, it means this component/variable is called multiple times in the compound system(with a cycle) in the same order as it appears in the gradient list.
   Ensure the feedback is aware of all sets of inputs and outputs.

{% if output_format_str %}
{{output_format_str}}
{% endif %}

<END_OF_SYSTEM_PROMPT>
<START_OF_USER>
<CONVERSATION>
{{conversation_sec}}
</CONVERSATION>
<OBJECTIVE_INSTRUCTION>
{{objective_instruction_sec}}
</OBJECTIVE_INSTRUCTION>
<END_OF_USER>
"""
# 6. If you receive error, must find one pred with error!
# 7. Ignore other metadata(noise such as id, data_id) in the data structure, only use the key(input) and key output that matters to infer the component functionality.

##############################################
# Loss Component
##############################################


OBJECTIVE_INSTRUCTION_BASE = r"""<OBJECTIVE_FUNCTION>
Your task is to provide the response with specific feedback based on the expected correct response (y_gt/ground_truth) and the score in the "<OUTPUTS/SCORE>".
Especially when the score is low.
Be CONCISE.

Be specific on why it has a low score.
Specify the difference between the expected correct response and the response.
</OBJECTIVE_FUNCTION>"""


OBJECTIVE_INSTRUCTION_CHAIN = r"""This conversation is part of a larger system. The <INPUTS/SCORE> was later used as "{{response_name}}: {{response_desc}}".
<OBJECTIVE_FUNCTION>
Your only goal is to clearly state how it obtained the "Eval output/score": {{response_gradient}}.
Especially when the score is low.
Be CONCISE.
If you have enough context, add more specific feedback on how it failed.
e.g. "The retrieved context is not enough to answer the question so the problem relies on the retrieval part."
</OBJECTIVE_FUNCTION>"""

###  Loss/Score Information  ###


# TODO: add predecessors awareness similar to CONVERSATION_START_INSTRUCTION_CHAIN
# used by GradComponent and loss component
LOSS_CONVERSATION_START_INSTRUCTION_STRING_FN = r"""
TARGET VARIABLE:
<NAME> {{variable.name}} </NAME>
<ROLE> {{variable.role_desc}} </ROLE>
<VARIABLE> {{variable.prompt_data}} </VARIABLE>
{{conversation_str}}
"""

# template to generate the conversation_str in loss component
LOSS_CONVERSATION_TEMPLATE_STRING = r"""
The variable is passed to the eval function and compared with a target/ground truth value to get
its score regarding to a SYSTEM_QUESTION: {{system_question}}.

EVAL_FUNC: {{eval_fn_desc}}

INPUTS to EVAL_FUNC:
{% for key, (value, eval_type) in inputs.items() %}
({{ key }}) (role: {{ value.role_desc }}),
data: {{ value.prompt_data }},
input_to_eval_fn: {{ value.eval_input }},
data_type: {{ eval_type }}
{% endfor %}

OUTPUTS/SCORE: {{response_value}}
{% if metadata %}
Note: {{metadata}}
{% endif %}"""

# template to generate the conversation_str
GRAD_COMPONENT_CONVERSATION_TEMPLATE_STRING = r"""
COMPONENT_DESC: {{component_desc}}

INPUTS:
{% for key, (value, eval_type) in inputs.items() %}
{{loop.index}}.
KEY: {{ key }}.
ROLE: {{ value.role_desc }},
DATA: {{ value.prompt_data }},
{% endfor %}

OUTPUT: {{response_value}}
{% if metadata %}
Note: {{metadata}}
{% endif %}"""


##############################################
# LLM as gradComponent
##############################################
# When the parameter has a gradient, it is the continuation of the backpropagation chain, a layer in the models
CONVERSATION_START_INSTRUCTION_CHAIN = r"""
{{variable_and_peers_info}}

{# system trainable variables #}
{% if predecessors %}
<START_OF_PREDECESSORS>
The target variable is used together with these predecessors variables besides of the peers:
{% for system_variable in predecessors %}
{{loop.index}}.
Name: {{system_variable.name}}
Type: {{system_variable.param_type}}
Description: {{system_variable.role_desc}}
WILL_BE_OPTIMIZED: {{system_variable.requires_opt}}
Value: {{system_variable.prompt_data}}
{% endfor %}
<END_OF_PREDECESSORS>
{% endif %}

Here is the inputs and output with this component(LM):
{{conversation_str}}
"""

# For the generator in the chain,
OBJECTIVE_INSTRUCTION_CHAIN = r"""
This component is part of a larger system. The <LM_OUTPUT> was later used as {{response_desc}}.
<OBJECTIVE_FUNCTION>
Your goal is to give feedback to the variable to guide the LLM_OUTPUT according to feedback: {{response_gradient}}
{% if instruction_to_backward_engine %}
Note: {{instruction_to_backward_engine}}
{% endif %}
</OBJECTIVE_FUNCTION>"""


VARIABLE_AND_PEERS_INFO = r"""
<START_OF_VARIABLE_DESC>
<NAME> {{variable.name}} </NAME>
<TYPE> {{variable.param_type}} </TYPE>
<ROLE> {{variable.role_desc}} </ROLE>
<VARIABLE>{{ variable.prompt_data}}</VARIABLE>
<END_OF_VARIABLE_DESC>
{% if peers %}
<VARIBLE_PEERS>
{% for peer in peers %}
{{loop.index}}.
PEER_NAME: {{peer.name}},
PEER_TYPE: {{peer.param_type}},
PEER_ROLE: {{peer.role_desc}}
WILL_BE_OPTIMIZED: {{peer.requires_opt}}
{% if peer.prompt_data %}
PEER_VARIABLE: {{peer.prompt_data}}
{% else %}
PEER_VARIABLE: EMPTY
{% endif %}
{% endfor %}
</VARIBLE_PEERS>
{% endif %}
"""

# The variable is used together with the these peer variables to instruct the language model on the task.
# - Do not overlap with the scope of the peer.


# a list of variables
ALL_PRED_INFO = r"""
<VARIABLES>
{% if variables %}
Length of the list: {{variables|length}}
{% for variable in variables %}
{{loop.index}}.
NAME: {{variable.name}},
TYPE: {{variable.param_type}},
ROLE: {{variable.role_desc}}
WILL_BE_OPTIMIZED: {{variable.requires_opt}}
VARIABLE: {{ variable.prompt_data}}
{% endfor %}
{% endif %}
</VARIABLES>
"""


###  Backward engine: user prompt
# First part to provide context of LLM as gradComponent
# The target variable is used as either input or a task instruction to a language model (LM):
# replace the "The target variable is used as either input or a task instruction to a language model (LM):" with the {{variable_desc}}
# NAME: {{variable_name}}
# Description: {{variable_desc}}
LLM_CONVERSATION_TEMPLATE = r"""
LM_INPUT: {{input_value}}
LM_OUTPUT: {{llm_output}}
{% if gt %}
GROUND_TRUTH: {{gt}}
{% endif %}
"""

# OUTPUT_INSTRUCTION = r"""
# You will create a feedback for each of the variable in the list above.
# If a variable will not be optimied, you just output empty string for that variable..
# NOTE: you MUST output a list of strings with the same length as the list above as ["...", "...", "..."]
# """
OUTPUT_INSTRUCTION = r"""
You will create a feedback for each of the variables in the list.
If a variable will not be optimized, you just output empty string.
Give enough details on the feedback.
Your output will be a list of strings with the SAME LENGTH as the <VARIABLES> list
as format of ["...", "...", "..."]
"""



"""Memory component for user-assistant conversations.

Memory can include data modeling, in-memory data storage, local file data storage, cloud data persistence, data pipeline, data retriever.
It is itself an LLM application and different use cases can do it differently.

This component handles the storage and retrieval of conversation history between users
and assistants. It provides local memory experience with the ability to format and
return conversation history.

Attributes:
    current_conversation (Conversation): Stores the current active conversation.
    turn_db (LocalDB): Database for storing all conversation turns.
    conver_db (LocalDB): Database for storing complete conversations.
"""

from uuid import uuid4
from adalflow.core.component import Component
from adalflow.core.db import LocalDB
from adalflow.core.types import (
    Conversation,
    DialogTurn,
    UserQuery,
    AssistantResponse,
)


class Memory(Component):
    def __init__(self, turn_db: LocalDB = None):
        """Initialize the Memory component.

        Args:
            turn_db (LocalDB, optional): Database for storing conversation turns.
                Defaults to None, in which case a new LocalDB is created.
        """
        super().__init__()
        self.current_conversation = Conversation()
        self.turn_db = turn_db or LocalDB()  # all turns
        self.conver_db = LocalDB()  # a list of conversations

    def call(self) -> str:
        """Returns the current conversation history as a formatted string.

        Returns:
            str: Formatted conversation history with alternating user and assistant messages.
                Returns empty string if no conversation history exists.
        """
        if not self.current_conversation.dialog_turns:
            return ""

        formatted_history = []
        for turn in self.current_conversation.dialog_turns.values():
            formatted_history.extend(
                [
                    f"User: {turn.user_query.query_str}",
                    f"Assistant: {turn.assistant_response.response_str}",
                ]
            )
        return "\n".join(formatted_history)

    def add_dialog_turn(self, user_query: str, assistant_response: str):
        """Add a new dialog turn to the current conversation.

        Args:
            user_query (str): The user's input message.
            assistant_response (str): The assistant's response message.
        """
        dialog_turn = DialogTurn(
            id=str(uuid4()),
            user_query=UserQuery(query_str=user_query),
            assistant_response=AssistantResponse(response_str=assistant_response),
        )

        self.current_conversation.append_dialog_turn(dialog_turn)

        self.turn_db.add(
            {"user_query": user_query, "assistant_response": assistant_response}
        )



import functools
import warnings
from typing import List, Optional, Dict, TYPE_CHECKING
import logging

if TYPE_CHECKING:
    from adalflow.core.generator import Generator
from adalflow.tracing import GeneratorStateLogger, GeneratorCallLogger

log = logging.getLogger(__name__)


def trace_generator_states(
    attributes: Optional[List[str]] = None,  # list of attributes of type Generator
    save_dir: Optional[str] = "./traces/",
    project_name: Optional[str] = None,
    filename: Optional[str] = None,
):
    r"""Decorator to trace generators in a task component.

    It dynamically attaches a GeneratorLogger to the target generator attribute and logs the prompt states of the generator.
    You can use it on any component that has attributes pointing to a generator object.

    Args:
        attributes (List[str], Optional): The list of attributes that point to the generator objects.
            If not provided, it will automatically detect the attributes that are instances of Generator.
        filepath (str, Optional): The path to the directory where the trace file will be saved. Default is "./traces/".
        filename (str, Optional): The name of the trace file. If not provided, it will be "{class_name}_generator_trace.json".

    Examples:

    .. code-block:: python

        from adalflow.tracing import trace_generator_states

        # Define a class and apply the decorator
        @trace_generator_states()
        class TestGenerator:
            def __init__(self):
                super().__init__()
                prompt_kwargs = {"input_str": "world"}
                self.generator = Generator(
                    model_client=OpenAIClient(),
                    prompt_kwargs=prompt_kwargs,
                    model_kwargs={"model": "gpt-3.5-turbo"},
                )
        # now you will see log files in the ./traces/ with a filename like TestGenerator_generator_trace.json
        # If you update the prompt templates or the prompt_kwargs, it will be logged in the file.
    """

    def decorator(cls):

        original_init = cls.__init__
        class_name = cls.__name__
        logger_project_name = project_name or class_name
        # final_filename = filename or f"{class_name}_generator_trace.json"
        # final_file = os.path.join(filepath, final_filename)

        # log.info(f"Tracing generator in {class_name} to {final_file}")

        @functools.wraps(original_init)
        def new_init(self, *args, **kwargs):
            from adalflow.core.generator import Generator

            original_init(self, *args, **kwargs)

            # enable automatic detection of generator attributes
            effective_attributes = attributes or [
                attr
                for attr in dir(self)
                if isinstance(getattr(self, attr, None), Generator)
            ]

            # create the logger in the current component
            if not hasattr(self, "generator_logger"):
                log.debug(f"Creating generator states logger for {class_name}")
                self.generator_logger = GeneratorStateLogger(
                    save_dir=save_dir,
                    project_name=logger_project_name,
                    filename=filename,
                )

            # Dynamically get the attribute to be logged if it exists.
            for attribute in effective_attributes:
                log.debug(f"Tracing generator states in {class_name}")
                target = getattr(self, attribute, None)
                generator_name = attribute

                if target is None:
                    warnings.warn(
                        f"Attribute {attribute} not found in {class_name}. Skipping tracing."
                    )
                    continue

                if not isinstance(target, Generator):
                    warnings.warn(
                        f"Attribute {attribute} is not a Generator instance. Skipping tracing."
                    )
                    continue

                # log the prompt states of the target generator
                self.generator_logger.log_prompt(target, generator_name)

        cls.__init__ = new_init
        return cls

    return decorator


def trace_generator_call(
    attributes: Optional[List[str]] = None,
    save_dir: Optional[str] = "./traces/",
    error_only: bool = True,
):
    r"""Decorator to trace generator predictions in a task component, especially failed ones.

    This decorator is a wrapper around the generator call method. It logs the generator call by
    reading its GeneratorOutput and logs the call if the output is an error.

    Args:
        attributes (List[str]): The list of attributes that point to the generator objects.
        save_dir (str): The path to the directory where the trace file will be saved.
        error_only (bool): If True, only log the calls that have an error. Default is True.

    Examples:

    .. code-block:: python

        from adalflow.tracing import trace_generator_call
        @trace_generator_call()
        class TestGenerator:
            def __init__(self):
                super().__init__()
                prompt_kwargs = {"input_str": "world"}
                self.generator = Generator(
                    model_client=OpenAIClient(),
                    prompt_kwargs=prompt_kwargs,
                    model_kwargs={"model": "gpt-3.5-turbo"},
                )
        # now you will see log files in the ./traces/ with a filename like TestGenerator_generator_call.jsonl
        # If the generator call has an error, it will be logged in the file.


    If you want to decorate a component(such as LLMRetriever) from the library where you do not have access to the source code, you can do it like this:

    .. code-block:: python

        from adalflow.components.retriever import LLMRetriever

        # Define a subclass and apply the decorator
        @trace_generator_call(save_dir=...)
        class LoggedLLMRetriever(LLMRetriever):
            pass
        retriever = LoggedLLMRetriever(...)


    You can access the logger via TestGenerator.generator_call_logger if you want to access call records in the code.
    """

    def decorator(cls):
        original_init = cls.__init__
        class_name = cls.__name__
        from adalflow.core.generator import Generator

        def _wrap_generator(
            generator_name: str,
            generator: "Generator",
            error_only: bool,
            logger: GeneratorCallLogger,
        ):
            r"""Wrap the call method of the generator to log the call."""
            original_call = generator.call

            @functools.wraps(original_call)
            def wrapped_call(*args, **kwargs):
                output = original_call(*args, **kwargs)
                try:
                    if (error_only and output.error is not None) or not error_only:
                        log.debug(f"Logging generator call for {generator_name}")
                        logger.log_call(
                            name=generator_name,
                            model_kwargs=kwargs.get("model_kwargs", {}),
                            prompt_kwargs=kwargs.get("prompt_kwargs", {}),
                            output=output,
                        )
                except Exception as e:
                    log.error(f"Error logging generator call for {generator_name}: {e}")
                return output

            return wrapped_call

        @functools.wraps(original_init)
        def new_init(self, *args, **kwargs):
            original_init(self, *args, **kwargs)

            # Find generator attributes
            effective_attributes = attributes or [
                attr
                for attr in dir(self)
                if isinstance(getattr(self, attr, None), Generator)
            ]
            generator_names_to_files: Dict[str, str] = {}
            # create the logger in the current component
            if not hasattr(self, "generator_call_logger"):
                self.generator_call_logger = GeneratorCallLogger(
                    save_dir=save_dir, project_name=class_name
                )

            generator_names_to_files = (
                self.generator_call_logger.generator_names_to_files
            )

            # Wrap each generator (with attr_name as the generator name)
            for attr_name in effective_attributes:
                target_generator = getattr(self, attr_name, None)
                if target_generator is None:
                    warnings.warn(
                        f"Attribute {attr_name} not found in {class_name}. Skipping tracing."
                    )
                    continue

                # handle the file registration
                if attr_name not in generator_names_to_files:

                    self.generator_call_logger.register_generator(attr_name)
                    filename = self.generator_call_logger.get_log_location(attr_name)
                    log.info(f"Registered generator {attr_name} with file {filename}")
                # Wrap the call method of the target generator
                if target_generator and hasattr(target_generator, "call"):
                    setattr(
                        target_generator,
                        "call",
                        _wrap_generator(
                            generator_name=attr_name,
                            generator=target_generator,
                            error_only=error_only,
                            logger=self.generator_call_logger,
                        ),
                    )

        cls.__init__ = new_init
        return cls

    return decorator



"""BM25 retriever implementation. """

from typing import List, Dict, Optional, Callable, Any, Sequence
import numpy as np
import heapq
import math
import logging

from adalflow.core.tokenizer import Tokenizer
from adalflow.core.types import (
    RetrieverOutput,
    RetrieverOutputType,
    RetrieverStrQueryType,
    RetrieverStrQueriesType,
    RetrieverDocumentsType,
)
from adalflow.core.retriever import (
    Retriever,
)
from adalflow.utils.file_io import save_json, load_json

log = logging.getLogger(__name__)

PARAM_K1 = 1.5
PARAM_B = 0.75
PARAM_EPSILON = 0.25


# TODO: move the functions in core.functional
def split_text_by_word_fn(x: str) -> List[str]:
    x = x.lower()
    return x.split(" ")


def split_text_by_word_fn_then_lower_tokenized(x: str) -> List[str]:
    tokenizer = Tokenizer()
    words = x.lower().split(" ")
    tokens = [tokenizer.encode(word) for word in words]
    final_tokens: List[str] = []
    for token_list in tokens:
        for token in token_list:
            final_tokens.append(tokenizer.decode([token]))
    return final_tokens


def split_text_tokenized(x: str) -> List[str]:
    tokenizer = Tokenizer()
    # words = x.lower().split(" ")
    tokens = tokenizer.encode(x)
    # print(tokens)
    final_tokens: List[str] = []
    for token in tokens:
        final_tokens.append(tokenizer.decode([token]))
    # print(final_tokens)
    return final_tokens


class BM25Retriever(Retriever[str, RetrieverStrQueryType]):
    __doc__ = r"""Fast Implementation of Best Matching 25 ranking function.

    It expects str as the final document type after ``document_map_func`` if the given document is not already in the format of List[str].
    It expects Union[str, Sequence[str]] as the input in :meth:`retrieve` method.

    .. math::

        \text{idf}(q_i) = \log\left(\frac{N - n(q_i) + 0.5}{n(q_i) + 0.5}\right)

        \text{score}(q, d) = \sum_{i=1}^{n} \text{idf}(q_i) \cdot \frac{f(q_i, d) \cdot (k1 + 1)}{f(q_i, d) + k1 \cdot \left(1 - b + b \cdot \frac{|d|}{\text{avgdl}}\right)}

    Explanation:
        - IDF(q_i) is the inverse document frequency of term q_i, which measures how important the term is. To avoid division by zero, 0.5 is added to the denominator, also for diminishing the weight of terms that occur very frequently in the document set and increase the weight of terms that occur rarely.
        - f(q_i, d) is the term frequency of term q_i in document d, which measures how often the term occurs in the document. The term frequency is normalized by dividing the raw term frequency by the document length.
        - |d| is the length of the document d in words or tokens.
        - avgdl is the average document length in the corpus.
        - N is the total number of documents in the corpus.
        - n(q_i) is the number of documents containing term q_i.

    References:
        [1] https://en.wikipedia.org/wiki/Okapi_BM25
        [2] https://github.com/dorianbrown/rank_bm25
        [3] Improvements to BM25 and Language Models Examined: https://www.cs.otago.ac.nz/homepages/andrew/papers/2014-2.pdf

    Args:
        top_k : (int): The number of documents to return
        k1 : (float, optional): Constant used for influencing the term frequency saturation. After saturation is reached, additional
                presence for the term adds a significantly less additional score. According to [1]_, experiments suggest
                that 1.2 < k1 < 2 yields reasonably good results, although the optimal value depends on factors such as
                the type of documents or queries.
        b : (float, optional): Constant used for influencing the effects of different document lengths relative to average document length.
                When b is bigger, lengthier documents (compared to average) have more impact on its effect. According to
                [1]_, experiments suggest that 0.5 < b < 0.8 yields reasonably good results, although the optimal value
                depends on factors such as the type of documents or queries.
        epsilon: (float, optional): Used to adapt the negative idf score to epilon * average_idf. Default is 0.25

        documents: (List[Any], optional): The list of documents to build the index from. Default is None.
        document_map_func: (Callable, optional): The function to transform the document into `List[str]`.
            You don't need it if your documents are already in format `List[str]`.
        use_tokenizer: (bool, optional): Whether to use the default tokenizer to split the text into words. Default is True.

    Examples:

    .. code-block:: python

            from adalflow.components.retriever.bm25_retriever import BM25Retriever

            documents = ["hello world", "world is beautiful", "today is a good day"]

    1. Pass the documents from the __init__ method:

    .. code-block:: python

            retriever = BM25Retriever(top_k=1, documents=documents)
            output = retriever("hello")
            print(output)
            # Output:
            # [RetrieverOutput(doc_indices=[0], doc_scores=[0.6229580777634034], query=None, documents=None)]

    2. Pass the documents from the :meth:`build_index_from_documents` method:

    .. code-block:: python
            retriever = BM25Retriever(top_k=1)
            retriever.build_index_from_documents(documents)
            output = retriever("hello")

    3. Save the index to file and load it back:

    .. code-block:: python

            retriever.save_to_file("bm25_index.json")
            retriever2 = BM25Retriever.load_from_file("bm25_index.json")
            output = retriever2("hello")
            print(output)

    note:
    The retriever only fill in the ``doc_indices`` and ``doc_scores``. The ``documents`` needs to be filled in by the user.
    """

    def __init__(
        self,
        top_k: int = 5,
        k1: float = PARAM_K1,
        b: float = PARAM_B,
        epsilon: float = PARAM_EPSILON,
        documents: Optional[Sequence[Any]] = None,
        document_map_func: Optional[Callable[[Any], str]] = None,
        use_tokenizer: bool = True,
    ):
        r"""
        - nd: <token, freq> (n(q_i) in the formula)
        - t2d: <token, <doc_index, freq>> (f(q_i, d) in the formula)
        - idf: <token, idf> (idf(q_i) in the formula)
        - doc_len: list of document lengths (|d| in the formula)
        - avgdl: average document length in the corpus (avgdl in the formula)
        - total_documents: total number of documents in the corpus (N in the formula)
        """
        super().__init__()
        self.k1 = k1
        self.b = b
        self.epsilon = epsilon
        self.top_k = top_k
        self._use_tokenizer = use_tokenizer
        self._split_function = (
            split_text_by_word_fn_then_lower_tokenized
            if use_tokenizer
            else split_text_by_word_fn
        )
        self.indexed = False  # this is important to check if the retrieve is possible
        self.index_keys = [
            "nd",
            "t2d",
            "idf",
            "doc_len",
            "avgdl",
            "total_documents",
            "top_k",
            "k1",
            "b",
            "epsilon",
            "indexed",
            "use_tokenizer",
        ]
        # initialize the index
        self.reset_index()
        self.documents = documents
        if documents:
            self.build_index_from_documents(documents, document_map_func)

    def reset_index(self):
        r"""Used for both initializing and resetting the index."""
        self.t2d: List[Dict[str, int]] = []  # term freuqency in each document
        self.nd: Dict[str, int] = {}  # number of documents containing the term
        self.idf: Dict[str, float] = {}  # idf of each term
        self.doc_len: List[int] = []  # list of document lengths
        self.avgdl: float = 0  # average document length
        self.indexed: bool = (
            False  # this is important to check if the retrieve is possible
        )
        self.total_documents: int = 0

    def _apply_split_function(self, documents: List[str]):
        if self._split_function is None:
            raise ValueError("split_function is not defined")
        if not documents:
            log.warning("No documents to split")
            return
        tokenized_documents = [self._split_function(doc) for doc in documents]

        return tokenized_documents

    def _initialize(self, corpus: List[List[str]]):
        r"""Initialize the term to document dictionary with the term frequencies in each document.
        The corpi is a list of tokenized documents."""

        self.total_documents = len(corpus)

        for document in corpus:
            self.doc_len.append(len(document))
            term_freq = {}
            for token in document:
                if token not in term_freq:
                    term_freq[token] = 0
                term_freq[token] += 1

            self.t2d.append(term_freq)

            for word, _ in term_freq.items():
                if word not in self.nd:
                    self.nd[word] = 0
                self.nd[word] += 1

        self.avgdl = sum(self.doc_len) / len(corpus)

    def _calc_idf(self):
        idf_sum = 0
        negative_idf = (
            []
        )  # idf can be negative if word is too common: more than half of the documents
        self.idf: Dict[str, float] = {}
        for token, freq in self.nd.items():
            idf = math.log(self.total_documents - freq + 0.5) - math.log(freq + 0.5)
            self.idf[token] = idf
            idf_sum += idf
            if idf < 0:
                negative_idf.append(token)
        self.average_idf = idf_sum / len(self.nd)  # average idf for each term

        # replace negative idf with epsilon * average_idf
        # NOTE: we can still have negative idf if most terms are too common, especially when the corpus is small
        eps = self.epsilon * self.average_idf
        for token in negative_idf:
            self.idf[token] = eps

    def _get_scores(self, query: List[str]) -> List[float]:
        r"""Calculate the BM25 score for the query and the documents in the corpus

        Args:
            query: List[str]: The tokenized query
        """
        score = np.zeros(self.total_documents)
        doc_len = np.array(self.doc_len)
        for q in query:
            q_freq = np.array([(doc.get(q) or 0) for doc in self.t2d])
            score += (self.idf.get(q) or 0) * (
                q_freq
                * (self.k1 + 1)
                / (q_freq + self.k1 * (1 - self.b + self.b * doc_len / self.avgdl))
            )
        return score.tolist()

    def _get_batch_scores(self, query: List[str], doc_ids: List[int]) -> List[float]:
        r"""Calculate the BM25 score for the query and the documents in the corpus

        Args:
            query: List[str]: The tokenized query
            doc_ids: List[int]: The list of document indexes to calculate the score
        """
        assert all(di < len(self.t2d) for di in doc_ids)
        score = np.zeros(len(doc_ids))
        doc_len = np.array(self.doc_len)[doc_ids]
        for q in query:
            q_freq = np.array([(self.t2d[di].get(q) or 0) for di in doc_ids])
            score += (self.idf.get(q) or 0) * (
                q_freq
                * (self.k1 + 1)
                / (q_freq + self.k1 * (1 - self.b + self.b * doc_len / self.avgdl))
            )
        return score.tolist()

    def build_index_from_documents(
        self,
        documents: RetrieverDocumentsType,
        document_map_func: Optional[Callable[[Any], str]] = None,
        **kwargs,
    ):
        r"""Built index from the `text` field of each document in the list of documents"""
        assert documents, "documents should not be empty"
        self.reset_index()
        self.documents = documents
        # the documents to be indexed
        if document_map_func:
            assert callable(document_map_func), "document_map_func should be callable"
            assert isinstance(
                document_map_func(documents[0]), str
            ), "document_map_func should return a string"
            list_of_documents_str = [document_map_func(doc) for doc in documents]
        else:
            list_of_documents_str = documents

        self.tokenized_documents = self._apply_split_function(list_of_documents_str)
        self._initialize(self.tokenized_documents)
        self._calc_idf()
        self.indexed = True

    def call(
        self, input: RetrieverStrQueriesType, top_k: Optional[int] = None, **kwargs
    ) -> RetrieverOutputType:
        """
        Retrieve the top n documents for the query and return only the indexes of the documents.

        Args:
            input: Union[str, List[str]]: The query or list of queries
            top_k: Optional[int]: The number of documents to return
        """
        if not self.indexed:
            raise ValueError("Index is not built. Please build the index first.")

        top_k = top_k or self.top_k
        output: RetrieverOutputType = []
        if isinstance(input, str):
            input = [input]
        elif isinstance(input, list):
            pass
        else:
            raise ValueError("input should be a string or a list of strings")
        # process each query
        for query in input:
            tokens = self._split_function(query)
            scores = self._get_scores(tokens)
            top_k_idx = heapq.nlargest(top_k, range(len(scores)), scores.__getitem__)
            top_k_scores = [scores[i] for i in top_k_idx]
            output.append(
                RetrieverOutput(
                    doc_indices=top_k_idx, doc_scores=top_k_scores, query=query
                )
            )
        return output

    def save_to_file(self, path: str):
        index_dict = super().to_dict()
        # filter out index_dict[data] that is not in self.index_keys
        for key in list(index_dict["data"].keys()):
            if key not in self.index_keys:
                del index_dict["data"][key]
        try:
            save_json(index_dict, path)
        except Exception as e:
            log.error(f"Error saving the index to file: {e}")
            raise e

    @classmethod
    def load_from_file(cls, path: str):
        # create an instance of the class
        try:
            index_dict = load_json(path)
            instance = cls.from_dict(index_dict)
            # add the split function
            instance._split_function = (
                split_text_by_word_fn_then_lower_tokenized
                if instance._use_tokenizer
                else split_text_by_word_fn
            )
            return instance
        except Exception as e:
            log.error(f"Error loading the index from file: {e}")
            raise e

    def _extra_repr(self) -> str:
        s = f"top_k={self.top_k}, k1={self.k1}, b={self.b}, epsilon={self.epsilon}, use_tokenizer={self._use_tokenizer}"
        s += f", total_documents={self.total_documents}"
        return s
